/*
 *  Power BI Visualizations
 *
 *  Copyright (c) Microsoft Corporation
 *  All rights reserved.
 *  MIT License
 *
 *  Permission is hereby granted, free of charge, to any person obtaining a copy
 *  of this software and associated documentation files (the ""Software""), to deal
 *  in the Software without restriction, including without limitation the rights
 *  to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 *  copies of the Software, and to permit persons to whom the Software is
 *  furnished to do so, subject to the following conditions:
 *
 *  The above copyright notice and this permission notice shall be included in
 *  all copies or substantial portions of the Software.
 *
 *  THE SOFTWARE IS PROVIDED *AS IS*, WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 *  IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 *  AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 *  OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 *  THE SOFTWARE.
 */

var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var EnhancedScatterChart1443994985041;
        (function (EnhancedScatterChart1443994985041) {
            var PixelConverter = jsCommon.PixelConverter;
            var createClassAndSelector = jsCommon.CssConstants.createClassAndSelector;
            var getCategoryIndexOfRole = powerbi.data.DataRoleHelper.getCategoryIndexOfRole;
            var getMeasureIndexOfRole = powerbi.data.DataRoleHelper.getMeasureIndexOfRole;
            var registerStandardInteractivityHandlers = powerbi.visuals.InteractivityUtils.registerStandardInteractivityHandlers;
            var Lazy = jsCommon.Lazy;
            var measureSvgTextWidth = powerbi.TextMeasurementService.measureSvgTextWidth;
            var measureSvgTextHeight = powerbi.TextMeasurementService.measureSvgTextHeight;
            var estimateSvgTextHeight = powerbi.TextMeasurementService.estimateSvgTextHeight;
            var svgEllipsis = powerbi.TextMeasurementService.svgEllipsis;
            var addTooltip = powerbi.visuals.TooltipManager.addTooltip;
            var fromPoint = jsCommon.PixelConverter.fromPoint;
            var LabelTextProperties = powerbi.visuals.dataLabelUtils.LabelTextProperties;
            var getLabelFormattedText = powerbi.visuals.dataLabelUtils.getLabelFormattedText;
            var equalWithPrecision = powerbi.Double.equalWithPrecision;
            var GetAnimationDuration = powerbi.visuals.AnimatorCommon.GetAnimationDuration;
            var getTailoredTextOrDefault = powerbi.TextMeasurementService.getTailoredTextOrDefault;
            var LegendData = powerbi.visuals.LegendData;
            var SelectionId = powerbi.visuals.SelectionId;
            var DataViewObject = powerbi.DataViewObject;
            var VisualDataRoleKind = powerbi.VisualDataRoleKind;
            var legendPosition = powerbi.visuals.legendPosition;
            var BaseAnimator = powerbi.visuals.BaseAnimator;
            var yAxisPosition = powerbi.visuals.yAxisPosition;
            var appendClearCatcher = powerbi.visuals.appendClearCatcher;
            var createInteractivityService = powerbi.visuals.createInteractivityService;
            var createLegend = powerbi.visuals.createLegend;
            var AxisHelper = powerbi.visuals.AxisHelper;
            var DataViewObjects = powerbi.DataViewObjects;
            var valueFormatter = powerbi.visuals.valueFormatter;
            var dataLabelUtils = powerbi.visuals.dataLabelUtils;
            var ColorHelper = powerbi.visuals.ColorHelper;
            var LegendIcon = powerbi.visuals.LegendIcon;
            var SelectionIdBuilder = powerbi.visuals.SelectionIdBuilder;
            var TooltipBuilder = powerbi.visuals.TooltipBuilder;
            var getInvalidValueWarnings = powerbi.visuals.getInvalidValueWarnings;
            var legendProps = powerbi.visuals.legendProps;
            var LegendPosition = powerbi.visuals.LegendPosition;
            var Legend = powerbi.visuals.Legend;
            var SVGUtil = powerbi.visuals.SVGUtil;
            var ObjectEnumerationBuilder = powerbi.visuals.ObjectEnumerationBuilder;
            var GradientUtils = powerbi.visuals.GradientUtils;
            var axisType = powerbi.visuals.axisType;
            var axisScale = powerbi.visuals.axisScale;
            var axisStyle = powerbi.visuals.axisStyle;
            var FontSize = powerbi.visuals.Units.FontSize;
            var EnhancedScatterChart = (function () {
                function EnhancedScatterChart() {
                    this.ScrollBarWidth = 10;
                }
                Object.defineProperty(EnhancedScatterChart.prototype, "margin", {
                    get: function () {
                        return this._margin || { left: 0, right: 0, top: 0, bottom: 0 };
                    },
                    set: function (value) {
                        this._margin = $.extend({}, value);
                        this._viewportIn = EnhancedScatterChart.substractMargin(this.viewport, this.margin);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(EnhancedScatterChart.prototype, "viewport", {
                    get: function () {
                        return this._viewport || { width: 0, height: 0 };
                    },
                    set: function (value) {
                        this._viewport = $.extend({}, value);
                        this._viewportIn = EnhancedScatterChart.substractMargin(this.viewport, this.margin);
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(EnhancedScatterChart.prototype, "viewportIn", {
                    get: function () {
                        return this._viewportIn || this.viewport;
                    },
                    enumerable: true,
                    configurable: true
                });
                Object.defineProperty(EnhancedScatterChart.prototype, "legendViewport", {
                    get: function () {
                        return this.legend.getMargins();
                    },
                    enumerable: true,
                    configurable: true
                });
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.getPropertiesByCapabilities = function (capabilities) {
                    var propertiesByCapabilities = {};
                    if (!capabilities) {
                        return propertiesByCapabilities;
                    }
                    for (var objectName in capabilities.objects) {
                        var objectDescriptor = capabilities.objects[objectName];
                        propertiesByCapabilities[objectName] = {};
                        if (objectDescriptor) {
                            for (var propertyName in objectDescriptor.properties) {
                                propertiesByCapabilities[objectName][propertyName] = {
                                    objectName: objectName,
                                    propertyName: propertyName
                                };
                            }
                        }
                    }
                    return propertiesByCapabilities;
                };
                EnhancedScatterChart.substractMargin = function (viewport, margin) {
                    return {
                        width: Math.max(viewport.width - (margin.left + margin.right), 0),
                        height: Math.max(viewport.height - (margin.top + margin.bottom), 0)
                    };
                };
                EnhancedScatterChart.getCustomSymbolType = function (shape) {
                    var customSymbolTypes = d3.map({
                        "circle": function (size) {
                            var r = Math.sqrt(size / Math.PI);
                            return "M0," + r + "A" + r + "," + r + " 0 1,1 0," + (-r) + "A" + r + "," + r + " 0 1,1 0," + r + "Z";
                        },
                        "cross": function (size) {
                            var r = Math.sqrt(size / 5) / 2;
                            return "M" + -3 * r + "," + -r
                                + "H" + -r + "V" + -3 * r + "H" + r + "V" + -r + "H" + 3 * r + "V" + r + "H" + r + "V" + 3 * r + "H" + -r + "V" + r + "H" + -3 * r + "Z";
                        },
                        "diamond": function (size) {
                            var ry = Math.sqrt(size / (2 * Math.tan(Math.PI / 6))), rx = ry * Math.tan(Math.PI / 6);
                            return "M0," + -ry
                                + "L" + rx + ",0"
                                + " 0," + ry
                                + " " + -rx + ",0"
                                + "Z";
                        },
                        "square": function (size) {
                            var r = Math.sqrt(size) / 2;
                            return "M" + -r + "," + -r
                                + "L" + r + "," + -r
                                + " " + r + "," + r
                                + " " + -r + "," + r
                                + "Z";
                        },
                        "triangle-up": function (size) {
                            var rx = Math.sqrt(size / Math.sqrt(3)), ry = rx * Math.sqrt(3) / 2;
                            return "M0," + -ry
                                + "L" + rx + "," + ry
                                + " " + -rx + "," + ry
                                + "Z";
                        },
                        "triangle-down": function (size) {
                            var rx = Math.sqrt(size / Math.sqrt(3)), ry = rx * Math.sqrt(3) / 2;
                            return "M0," + ry
                                + "L" + rx + "," + -ry
                                + " " + -rx + "," + -ry
                                + "Z";
                        },
                        "star": function (size) {
                            var outerRadius = Math.sqrt(size / 2);
                            var innerRadius = Math.sqrt(size / 10);
                            var results = "";
                            var angle = Math.PI / 5;
                            for (var i = 0; i < 10; i++) {
                                // Use outer or inner radius depending on what iteration we are in.
                                var r = (i & 1) === 0 ? outerRadius : innerRadius;
                                var currX = Math.cos(i * angle) * r;
                                var currY = Math.sin(i * angle) * r;
                                // Our first time we simply append the coordinates, subsequet times
                                // we append a ", " to distinguish each coordinate pair.
                                if (i === 0) {
                                    results = "M" + currX + "," + currY + "L";
                                }
                                else {
                                    results += " " + currX + "," + currY;
                                }
                            }
                            return results + "Z";
                        },
                        "hexagon": function (size) {
                            var r = Math.sqrt(size / (6 * Math.sqrt(3)));
                            var r2 = Math.sqrt(size / (2 * Math.sqrt(3)));
                            return "M0," + (2 * r) + "L" + (-r2) + "," + r + " " + (-r2) + "," + (-r) + " 0," + (-2 * r) + " " + r2 + "," + (-r) + " " + r2 + "," + r + "Z";
                        },
                        "x": function (size) {
                            var r = Math.sqrt(size / 10);
                            return "M0," + r + "L" + (-r) + "," + 2 * r + " " + (-2 * r) + "," + r + " " + (-r) + ",0 " + (-2 * r) + "," + (-r) + " " + (-r) + "," + (-2 * r) + " 0," + (-r) + " " + r + "," + (-2 * r) + " " + (2 * r) + "," + (-r) + " " + r + ",0 " + (2 * r) + "," + r + " " + r + "," + (2 * r) + "Z";
                        },
                        "uparrow": function (size) {
                            var r = Math.sqrt(size / 12);
                            return "M" + r + "," + (3 * r) + "L" + (-r) + "," + (3 * r) + " " + (-r) + "," + (-r) + " " + (-2 * r) + "," + (-r) + " 0," + (-3 * r) + " " + (2 * r) + "," + (-r) + " " + r + "," + (-r) + "Z";
                        },
                        "downarrow": function (size) {
                            var r = Math.sqrt(size / 12);
                            return "M0," + (3 * r) + "L" + (-2 * r) + "," + r + " " + (-r) + "," + r + " " + (-r) + "," + (-3 * r) + " " + r + "," + (-3 * r) + " " + r + "," + r + " " + (2 * r) + "," + r + "Z";
                        }
                    });
                    var defaultValue = customSymbolTypes.entries()[0].value;
                    if (!shape) {
                        return defaultValue;
                    }
                    else if (isNaN(shape)) {
                        return customSymbolTypes[shape && shape.toString().toLowerCase()] || defaultValue;
                    }
                    else {
                        var result = customSymbolTypes.entries()[Math.floor(shape)];
                        return result ? result.value : defaultValue;
                    }
                };
                EnhancedScatterChart.prototype.init = function (options) {
                    this.options = options;
                    this.animator = new BaseAnimator();
                    this.behavior = new CustomVisualBehavior([new EnhancedScatterChartWebBehavior(EnhancedScatterChart.DimmedBubbleOpacity, EnhancedScatterChart.DefaultBubbleOpacity)]);
                    var element = this.element = options.element;
                    this.viewport = _.clone(options.viewport);
                    this.style = options.style;
                    this.hostServices = options.host;
                    this.colors = this.style.colorPalette.dataColors;
                    this.interactivity = options.interactivity;
                    this.margin = {
                        top: 1,
                        right: 1,
                        bottom: 1,
                        left: 1
                    };
                    this.yAxisOrientation = yAxisPosition.left;
                    this.adjustMargins();
                    var showLinesOnX = this.scrollY = true;
                    var showLinesOnY = this.scrollX = true;
                    var svg = this.svg = d3.select(element.get(0))
                        .append("svg")
                        .style("position", "absolute")
                        .classed(EnhancedScatterChart.ClassName, true);
                    var axisGraphicsContext = this.axisGraphicsContext = svg.append("g")
                        .classed(EnhancedScatterChart.AxisGraphicsContextClassName, true);
                    this.svgScrollable = svg.append("svg")
                        .classed("svgScrollable", true)
                        .style("overflow", "hidden");
                    var axisGraphicsContextScrollable = this.axisGraphicsContextScrollable = this.svgScrollable.append("g")
                        .classed(EnhancedScatterChart.AxisGraphicsContextClassName, true);
                    this.clearCatcher = appendClearCatcher(this.axisGraphicsContextScrollable);
                    var axisGroup = showLinesOnX
                        ? axisGraphicsContextScrollable
                        : axisGraphicsContext;
                    this.backgroundGraphicsContext = axisGraphicsContext.append("svg:image");
                    this.xAxisGraphicsContext = showLinesOnX
                        ? axisGraphicsContext.append("g").attr("class", "x axis")
                        : axisGraphicsContextScrollable.append("g").attr("class", "x axis");
                    this.y1AxisGraphicsContext = axisGroup.append("g").attr("class", "y axis");
                    this.xAxisGraphicsContext.classed("showLinesOnAxis", showLinesOnX);
                    this.y1AxisGraphicsContext.classed("showLinesOnAxis", showLinesOnY);
                    this.xAxisGraphicsContext.classed("hideLinesOnAxis", !showLinesOnX);
                    this.y1AxisGraphicsContext.classed("hideLinesOnAxis", !showLinesOnY);
                    this.interactivityService = createInteractivityService(this.hostServices);
                    this.legend = createLegend(element, this.interactivity && this.interactivity.isInteractiveLegend, this.interactivityService, true);
                    this.mainGraphicsG = this.axisGraphicsContextScrollable
                        .append("g")
                        .classed(EnhancedScatterChart.MainGraphicsContextClassName, true);
                    this.mainGraphicsSVGSelection = this.mainGraphicsG.append("svg");
                    this.mainGraphicsContext = this.mainGraphicsSVGSelection.append("g");
                    this.svgDefaultImage = "";
                    this.keyArray = [];
                };
                EnhancedScatterChart.prototype.adjustMargins = function () {
                    // Adjust margins if ticks are not going to be shown on either axis
                    var xAxis = this.element.find(".x.axis");
                    if (AxisHelper.getRecommendedNumberOfTicksForXAxis(this.viewportIn.width) === 0
                        && AxisHelper.getRecommendedNumberOfTicksForYAxis(this.viewportIn.height) === 0) {
                        this.margin = {
                            top: 0,
                            right: 0,
                            bottom: 0,
                            left: 0
                        };
                        xAxis.hide();
                    }
                    else {
                        xAxis.show();
                    }
                };
                EnhancedScatterChart.prototype.getValueAxisProperties = function (dataViewMetadata, axisTitleOnByDefault) {
                    var toReturn = {};
                    if (!dataViewMetadata) {
                        return toReturn;
                    }
                    var objects = dataViewMetadata.objects;
                    if (objects) {
                        var valueAxisObject = objects["valueAxis"];
                        if (valueAxisObject) {
                            toReturn = {
                                show: valueAxisObject["show"],
                                position: valueAxisObject["position"],
                                axisScale: valueAxisObject["axisScale"],
                                start: valueAxisObject["start"],
                                end: valueAxisObject["end"],
                                showAxisTitle: valueAxisObject["showAxisTitle"] == null ? axisTitleOnByDefault : valueAxisObject["showAxisTitle"],
                                axisStyle: valueAxisObject["axisStyle"],
                                axisColor: valueAxisObject["axisColor"],
                                secShow: valueAxisObject["secShow"],
                                secPosition: valueAxisObject["secPosition"],
                                secAxisScale: valueAxisObject["secAxisScale"],
                                secStart: valueAxisObject["secStart"],
                                secEnd: valueAxisObject["secEnd"],
                                secShowAxisTitle: valueAxisObject["secShowAxisTitle"],
                                secAxisStyle: valueAxisObject["secAxisStyle"],
                                labelDisplayUnits: valueAxisObject["labelDisplayUnits"],
                            };
                        }
                    }
                    return toReturn;
                };
                EnhancedScatterChart.prototype.getCategoryAxisProperties = function (dataViewMetadata, axisTitleOnByDefault) {
                    var toReturn = {};
                    if (!dataViewMetadata) {
                        return toReturn;
                    }
                    var objects = dataViewMetadata.objects;
                    if (objects) {
                        var categoryAxisObject = objects["categoryAxis"];
                        if (categoryAxisObject) {
                            toReturn = {
                                show: categoryAxisObject["show"],
                                axisType: categoryAxisObject["axisType"],
                                axisScale: categoryAxisObject["axisScale"],
                                axisColor: categoryAxisObject["axisColor"],
                                start: categoryAxisObject["start"],
                                end: categoryAxisObject["end"],
                                showAxisTitle: categoryAxisObject["showAxisTitle"] == null
                                    ? axisTitleOnByDefault : categoryAxisObject["showAxisTitle"],
                                axisStyle: categoryAxisObject["axisStyle"],
                                labelDisplayUnits: categoryAxisObject["labelDisplayUnits"]
                            };
                        }
                    }
                    return toReturn;
                };
                EnhancedScatterChart.converter = function (dataView, colorPalette, interactivityService, categoryAxisProperties, valueAxisProperties) {
                    if (!dataView) {
                        return EnhancedScatterChart.getDefaultData();
                    }
                    var categoryValues, categoryFormatter, categoryObjects, categoryIdentities, categoryQueryName, dataViewCategorical = dataView.categorical, dataViewMetadata = dataView.metadata, categories = dataViewCategorical.categories || [], dataValues = dataViewCategorical.values, hasDynamicSeries = !!dataValues.source, grouped = dataValues.grouped(), dvSource = dataValues.source, scatterMetadata = EnhancedScatterChart.getMetadata(categories, grouped, dvSource), categoryIndex = scatterMetadata.idx.category, useShape = scatterMetadata.idx.image >= 0, useCustomColor = scatterMetadata.idx.colorFill >= 0;
                    if (dataViewCategorical.categories &&
                        dataViewCategorical.categories.length > 0 &&
                        dataViewCategorical.categories[categoryIndex]) {
                        var mainCategory = dataViewCategorical.categories[categoryIndex];
                        categoryValues = mainCategory.values;
                        categoryFormatter = valueFormatter.create({
                            format: valueFormatter.getFormatString(mainCategory.source, EnhancedScatterChart.Properties["general"]["formatString"]),
                            value: categoryValues[0],
                            value2: categoryValues[categoryValues.length - 1],
                            columnType: mainCategory.source.type,
                        });
                        categoryIdentities = mainCategory.identity;
                        categoryObjects = mainCategory.objects;
                        categoryQueryName = mainCategory.source ? mainCategory.source.queryName : null;
                    }
                    else {
                        categoryValues = [null];
                        // creating default formatter for null value (to get the right string of empty value from the locale)
                        categoryFormatter = valueFormatter.createDefaultFormatter(null);
                    }
                    var dataLabelsSettings = dataLabelUtils.getDefaultPointLabelSettings(), fillPoint = false, backdrop = { show: false, url: "" }, crosshair = false, outline = false, defaultDataPointColor = "", showAllDataPoints = true;
                    if (dataViewMetadata && dataViewMetadata.objects) {
                        var objects = dataViewMetadata.objects;
                        defaultDataPointColor = DataViewObjects.getFillColor(objects, EnhancedScatterChart.Properties["dataPoint"]["defaultColor"]);
                        showAllDataPoints = DataViewObjects.getValue(objects, EnhancedScatterChart.Properties["dataPoint"]["showAllDataPoints"]);
                        var labelsObj = objects["categoryLabels"];
                        if (labelsObj) {
                            dataLabelsSettings.show = (labelsObj["show"] !== undefined)
                                ? labelsObj["show"] : dataLabelsSettings.show;
                            dataLabelsSettings.fontSize = (labelsObj["fontSize"] !== undefined)
                                ? labelsObj["fontSize"] : dataLabelsSettings.fontSize;
                            if (labelsObj["color"] !== undefined) {
                                dataLabelsSettings.labelColor = labelsObj["color"].solid.color;
                            }
                        }
                        fillPoint = DataViewObjects.getValue(objects, EnhancedScatterChart.Properties["fillPoint"]["show"], fillPoint);
                        var backdropObject = objects["backdrop"];
                        if (backdropObject !== undefined) {
                            backdrop.show = backdropObject["show"];
                            if (backdrop.show) {
                                backdrop.url = backdropObject["url"];
                            }
                        }
                        var crosshairObject = objects["crosshair"];
                        if (crosshairObject !== undefined) {
                            crosshair = crosshairObject["show"];
                        }
                        var outlineObject = objects["outline"];
                        if (outlineObject !== undefined) {
                            outline = outlineObject["show"];
                        }
                    }
                    var dataPoints = EnhancedScatterChart.createDataPoints(dataValues, scatterMetadata, categories, categoryValues, categoryFormatter, categoryIdentities, categoryObjects, colorPalette, hasDynamicSeries, dataLabelsSettings, defaultDataPointColor, categoryQueryName);
                    if (interactivityService) {
                        interactivityService.applySelectionStateToData(dataPoints);
                    }
                    var legendItems = [];
                    if (hasDynamicSeries) {
                        var formatString = valueFormatter.getFormatString(dvSource, EnhancedScatterChart.Properties["general"]["formatString"]);
                        legendItems = EnhancedScatterChart.createSeriesLegend(dataValues, colorPalette, dataValues, formatString, defaultDataPointColor);
                    }
                    var legendTitle = dataValues && dvSource
                        ? dvSource.displayName
                        : "";
                    if (!legendTitle) {
                        legendTitle = categories &&
                            categories[categoryIndex] &&
                            categories[categoryIndex].source &&
                            categories[categoryIndex].source.displayName
                            ? categories[categoryIndex].source.displayName : "";
                    }
                    var legendData = { title: legendTitle, dataPoints: legendItems };
                    var sizeRange = EnhancedScatterChart.getSizeRangeForGroups(grouped, scatterMetadata.idx.size);
                    if (categoryAxisProperties && categoryAxisProperties["showAxisTitle"] !== null && categoryAxisProperties["showAxisTitle"] === false) {
                        scatterMetadata.axesLabels.x = null;
                    }
                    if (valueAxisProperties && valueAxisProperties["showAxisTitle"] !== null && valueAxisProperties["showAxisTitle"] === false) {
                        scatterMetadata.axesLabels.y = null;
                    }
                    if (dataPoints && dataPoints[0]) {
                        var point = dataPoints[0];
                        if (point.backdrop != null) {
                            backdrop.show = true;
                            backdrop.url = point.backdrop;
                        }
                        if (point.xStart != null) {
                            categoryAxisProperties["start"] = point.xStart;
                        }
                        if (point.xEnd != null) {
                            categoryAxisProperties["end"] = point.xEnd;
                        }
                        if (point.yStart != null) {
                            valueAxisProperties["start"] = point.yStart;
                        }
                        if (point.yEnd != null) {
                            valueAxisProperties["end"] = point.yEnd;
                        }
                    }
                    return {
                        xCol: scatterMetadata.cols.x,
                        yCol: scatterMetadata.cols.y,
                        dataPoints: dataPoints,
                        legendData: legendData,
                        axesLabels: scatterMetadata.axesLabels,
                        selectedIds: [],
                        size: scatterMetadata.cols.size,
                        sizeRange: sizeRange,
                        dataLabelsSettings: dataLabelsSettings,
                        defaultDataPointColor: defaultDataPointColor,
                        hasDynamicSeries: hasDynamicSeries,
                        showAllDataPoints: showAllDataPoints,
                        fillPoint: fillPoint,
                        useShape: useShape,
                        useCustomColor: useCustomColor,
                        backdrop: backdrop,
                        crosshair: crosshair,
                        outline: outline
                    };
                };
                EnhancedScatterChart.createSeriesLegend = function (dataValues, colorPalette, categorical, formatString, defaultDataPointColor) {
                    var legendItems = [], grouped = dataValues.grouped(), colorHelper = new ColorHelper(colorPalette, EnhancedScatterChart.Properties["dataPoint"]["fill"], defaultDataPointColor);
                    for (var i = 0, len = grouped.length; i < len; i++) {
                        var grouping = grouped[i], selectionId, color;
                        color = colorHelper.getColorForSeriesValue(grouping.objects, dataValues.identityFields, grouping.name);
                        selectionId = grouping.identity
                            ? SelectionId.createWithId(grouping.identity)
                            : SelectionId.createNull();
                        legendItems.push({
                            color: color,
                            icon: LegendIcon.Circle,
                            label: valueFormatter.format(grouping.name, formatString),
                            identity: selectionId,
                            selected: false,
                        });
                    }
                    return legendItems;
                };
                EnhancedScatterChart.getSizeRangeForGroups = function (dataViewValueGroups, sizeColumnIndex) {
                    var result = {};
                    if (dataViewValueGroups) {
                        dataViewValueGroups.forEach(function (group) {
                            var sizeColumn = EnhancedScatterChart.getMeasureValue(sizeColumnIndex, group.values), currentRange = AxisHelper.getRangeForColumn(sizeColumn);
                            if (result.min == null || result.min > currentRange.min) {
                                result.min = currentRange.min;
                            }
                            if (result.max == null || result.max < currentRange.max) {
                                result.max = currentRange.max;
                            }
                        });
                    }
                    return result;
                };
                EnhancedScatterChart.getMetadata = function (categories, grouped, source) {
                    var categoryIndex = getCategoryIndexOfRole(categories, EnhancedScatterChart.ColumnCategory), colorFillIndex = getCategoryIndexOfRole(categories, EnhancedScatterChart.ColumnColorFill), imageIndex = getCategoryIndexOfRole(categories, EnhancedScatterChart.ColumnImage), backdropIndex = getCategoryIndexOfRole(categories, EnhancedScatterChart.ColumnBackdrop), xIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnX), yIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnY), sizeIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnSize), shapeIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnShape), rotationIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnRotation), xStartIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnXStart), xEndIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnXEnd), yStartIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnYStart), yEndIndex = getMeasureIndexOfRole(grouped, EnhancedScatterChart.ColumnYEnd), xCol, yCol, sizeCol, xAxisLabel = "", yAxisLabel = "";
                    if (grouped && grouped.length) {
                        var firstGroup = grouped[0];
                        if (xIndex >= 0) {
                            xCol = firstGroup.values[xIndex].source;
                            xAxisLabel = firstGroup.values[xIndex].source.displayName;
                        }
                        if (yIndex >= 0) {
                            yCol = firstGroup.values[yIndex].source;
                            yAxisLabel = firstGroup.values[yIndex].source.displayName;
                        }
                        if (sizeIndex >= 0) {
                            sizeCol = firstGroup.values[sizeIndex].source;
                        }
                    }
                    return {
                        idx: {
                            category: categoryIndex,
                            x: xIndex,
                            y: yIndex,
                            size: sizeIndex,
                            colorFill: colorFillIndex,
                            shape: shapeIndex,
                            image: imageIndex,
                            rotation: rotationIndex,
                            backdrop: backdropIndex,
                            xStart: xStartIndex,
                            xEnd: xEndIndex,
                            yStart: yStartIndex,
                            yEnd: yEndIndex
                        },
                        cols: {
                            x: xCol,
                            y: yCol,
                            size: sizeCol
                        },
                        axesLabels: {
                            x: xAxisLabel,
                            y: yAxisLabel
                        }
                    };
                };
                EnhancedScatterChart.createLazyFormattedCategory = function (formatter, value) {
                    return new Lazy(function () { return formatter.format(value); });
                };
                EnhancedScatterChart.createDataPoints = function (dataValues, metadata, categories, categoryValues, categoryFormatter, categoryIdentities, categoryObjects, colorPalette, hasDynamicSeries, labelSettings, defaultDataPointColor, categoryQueryName) {
                    var dataPoints = [], colorHelper, indicies = metadata.idx, formatStringProp, dataValueSource = dataValues.source, grouped = dataValues.grouped(), fontSizeInPx = PixelConverter.fromPoint(labelSettings.fontSize);
                    formatStringProp = EnhancedScatterChart.Properties["general"]["formatString"];
                    colorHelper = new ColorHelper(colorPalette, EnhancedScatterChart.Properties["dataPoint"]["fill"], defaultDataPointColor);
                    for (var categoryIdx = 0, ilen = categoryValues.length; categoryIdx < ilen; categoryIdx++) {
                        var categoryValue = categoryValues[categoryIdx];
                        for (var seriesIdx = 0, len = grouped.length; seriesIdx < len; seriesIdx++) {
                            var measureColorFill = categories[indicies.colorFill], measureImage = categories[indicies.image], measureBackdrop = categories[indicies.backdrop];
                            var grouping = grouped[seriesIdx], seriesValues = grouping.values, measureX = EnhancedScatterChart.getMeasureValue(indicies.x, seriesValues), measureY = EnhancedScatterChart.getMeasureValue(indicies.y, seriesValues), measureSize = EnhancedScatterChart.getMeasureValue(indicies.size, seriesValues), measureShape = EnhancedScatterChart.getMeasureValue(indicies.shape, seriesValues), measureRotation = EnhancedScatterChart.getMeasureValue(indicies.rotation, seriesValues), measureXStart = EnhancedScatterChart.getMeasureValue(indicies.xStart, seriesValues), measureXEnd = EnhancedScatterChart.getMeasureValue(indicies.xEnd, seriesValues), measureYStart = EnhancedScatterChart.getMeasureValue(indicies.yStart, seriesValues), measureYEnd = EnhancedScatterChart.getMeasureValue(indicies.yEnd, seriesValues);
                            //TODO: need to update (refactor) these lines below.
                            var xVal = measureX && measureX.values && !isNaN(measureX.values[categoryIdx]) ? measureX.values[categoryIdx] : null, yVal = measureY && measureY.values && !isNaN(measureY.values[categoryIdx]) ? measureY.values[categoryIdx] : 0;
                            var hasNullValue = (xVal == null) || (yVal == null);
                            if (hasNullValue) {
                                continue;
                            }
                            var size, colorFill, shapeSymbolType, image, rotation, backdrop, xStart, xEnd, yStart, yEnd, color;
                            size = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureSize, categoryIdx);
                            colorFill = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureColorFill, categoryIdx);
                            shapeSymbolType = EnhancedScatterChart.getCustomSymbolType(EnhancedScatterChart.getValueFromDataViewValueColumnById(measureShape, categoryIdx));
                            image = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureImage, categoryIdx);
                            rotation = EnhancedScatterChart.getNumberFromDataViewValueColumnById(measureRotation, categoryIdx);
                            backdrop = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureBackdrop, categoryIdx);
                            xStart = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureXStart, categoryIdx);
                            xEnd = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureXEnd, categoryIdx);
                            yStart = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureYStart, categoryIdx);
                            yEnd = EnhancedScatterChart.getValueFromDataViewValueColumnById(measureYEnd, categoryIdx);
                            if (hasDynamicSeries) {
                                color = colorHelper.getColorForSeriesValue(grouping.objects, dataValues.identityFields, grouping.name);
                            }
                            else {
                                // If we have no Size measure then use a blank query name
                                var measureSource = (measureSize != null)
                                    ? measureSize.source.queryName
                                    : "";
                                color = colorHelper.getColorForMeasure(categoryObjects && categoryObjects[categoryIdx], measureSource);
                            }
                            var category = categories && categories.length > 0
                                ? categories[indicies.category]
                                : null;
                            var identity = SelectionIdBuilder.builder()
                                .withCategory(category, categoryIdx)
                                .withSeries(dataValues, grouping)
                                .createSelectionId();
                            //TODO: need to refactor these lines below.
                            var seriesData = [];
                            if (dataValueSource) {
                                // Dynamic series
                                seriesData.push({ value: grouping.name, metadata: { source: dataValueSource, values: [] } });
                            }
                            if (measureX) {
                                seriesData.push({ value: xVal, metadata: measureX });
                            }
                            if (measureY) {
                                seriesData.push({ value: yVal, metadata: measureY });
                            }
                            if (measureSize && measureSize.values && measureSize.values.length > 0) {
                                seriesData.push({ value: measureSize.values[categoryIdx], metadata: measureSize });
                            }
                            if (measureColorFill && measureColorFill.values && measureColorFill.values.length > 0) {
                                seriesData.push({ value: measureColorFill.values[categoryIdx], metadata: measureColorFill });
                            }
                            if (measureShape && measureShape.values && measureShape.values.length > 0) {
                                seriesData.push({ value: measureShape.values[categoryIdx], metadata: measureShape });
                            }
                            if (measureImage && measureImage.values && measureImage.values.length > 0) {
                                seriesData.push({ value: measureImage.values[categoryIdx], metadata: measureImage });
                            }
                            if (measureRotation && measureRotation.values && measureRotation.values.length > 0) {
                                seriesData.push({ value: measureRotation.values[categoryIdx], metadata: measureRotation });
                            }
                            if (measureBackdrop && measureBackdrop.values && measureBackdrop.values.length > 0) {
                                seriesData.push({ value: measureBackdrop.values[categoryIdx], metadata: measureBackdrop });
                            }
                            if (measureXStart && measureXStart.values && measureXStart.values.length > 0) {
                                seriesData.push({ value: measureXStart.values[categoryIdx], metadata: measureXStart });
                            }
                            if (measureXEnd && measureXEnd.values && measureXEnd.values.length > 0) {
                                seriesData.push({ value: measureXEnd.values[categoryIdx], metadata: measureXEnd });
                            }
                            if (measureYStart && measureYStart.values && measureYStart.values.length > 0) {
                                seriesData.push({ value: measureYStart.values[categoryIdx], metadata: measureYStart });
                            }
                            if (measureYEnd && measureYEnd.values && measureYEnd.values.length > 0) {
                                seriesData.push({ value: measureYEnd.values[categoryIdx], metadata: measureYEnd });
                            }
                            var tooltipInfo = TooltipBuilder.createTooltipInfo(formatStringProp, /* formatStringProp */ undefined, /* dataViewCat */ categoryValue, /* categoryValue */ null, /* value */ category ? [category] : undefined, /* categories */ seriesData, /* seriesData */ undefined /* seriesIndex */);
                            var dataPoint = {
                                x: xVal,
                                y: yVal,
                                size: size,
                                radius: { sizeMeasure: measureSize, index: categoryIdx },
                                fill: color,
                                formattedCategory: this.createLazyFormattedCategory(categoryFormatter, categoryValue),
                                selected: false,
                                identity: identity,
                                tooltipInfo: tooltipInfo,
                                labelFill: labelSettings.labelColor,
                                labelFontSize: fontSizeInPx,
                                contentPosition: 8,
                                colorFill: colorFill,
                                shapeSymbolType: shapeSymbolType,
                                svgurl: image,
                                rotation: rotation,
                                backdrop: backdrop,
                                xStart: xStart,
                                xEnd: xEnd,
                                yStart: yStart,
                                yEnd: yEnd
                            };
                            dataPoints.push(dataPoint);
                        }
                    }
                    return dataPoints;
                };
                EnhancedScatterChart.getMeasureValue = function (measureIndex, seriesValues) {
                    if (seriesValues && measureIndex >= 0) {
                        return seriesValues[measureIndex];
                    }
                    return null;
                };
                EnhancedScatterChart.getNumberFromDataViewValueColumnById = function (dataViewValueColumn, index) {
                    var value = EnhancedScatterChart.getValueFromDataViewValueColumnById(dataViewValueColumn, index);
                    return value && !isNaN(value)
                        ? value
                        : 0;
                };
                EnhancedScatterChart.getValueFromDataViewValueColumnById = function (dataViewValueColumn, index) {
                    return dataViewValueColumn && dataViewValueColumn.values
                        ? dataViewValueColumn.values[index]
                        : null;
                };
                EnhancedScatterChart.getDefaultData = function () {
                    return {
                        xCol: undefined,
                        yCol: undefined,
                        dataPoints: [],
                        legendData: { dataPoints: [] },
                        axesLabels: { x: "", y: "" },
                        selectedIds: [],
                        sizeRange: [],
                        dataLabelsSettings: dataLabelUtils.getDefaultPointLabelSettings(),
                        defaultDataPointColor: null,
                        hasDynamicSeries: false,
                        useShape: false,
                        useCustomColor: false,
                    };
                };
                EnhancedScatterChart.prototype.setData = function (dataViews) {
                    this.data = EnhancedScatterChart.getDefaultData();
                    if (dataViews && dataViews.length > 0) {
                        var dataView = dataViews[0];
                        if (dataView) {
                            this.categoryAxisProperties = this.getCategoryAxisProperties(dataView.metadata, true);
                            this.valueAxisProperties = this.getValueAxisProperties(dataView.metadata, true);
                            this.dataView = dataView;
                            if (dataView.categorical && dataView.categorical.values) {
                                this.data = EnhancedScatterChart.converter(dataView, this.colors, this.interactivityService, this.categoryAxisProperties, this.valueAxisProperties);
                            }
                        }
                    }
                };
                EnhancedScatterChart.prototype.update = function (options) {
                    debug.assertValue(options, "options");
                    var dataViews = this.dataViews = options.dataViews;
                    this.viewport = _.clone(options.viewport);
                    if (!dataViews) {
                        return;
                    }
                    if (dataViews && dataViews.length > 0) {
                        var warnings = getInvalidValueWarnings(dataViews, false /*supportsNaN*/, false /*supportsNegativeInfinity*/, false /*supportsPositiveInfinity*/);
                        if (warnings && warnings.length > 0)
                            this.hostServices.setWarnings(warnings);
                        this.populateObjectProperties(dataViews);
                    }
                    this.setData(dataViews);
                    // Note: interactive legend shouldn"t be rendered explicitly here
                    // The interactive legend is being rendered in the render method of ICartesianVisual
                    if (!(this.options.interactivity && this.options.interactivity.isInteractiveLegend)) {
                        this.renderLegend();
                    }
                    this.render(options.suppressAnimations);
                };
                EnhancedScatterChart.prototype.populateObjectProperties = function (dataViews) {
                    if (dataViews && dataViews.length > 0) {
                        var dataViewMetadata = dataViews[0].metadata;
                        if (dataViewMetadata) {
                            this.legendObjectProperties = DataViewObjects.getObject(dataViewMetadata.objects, "legend", {});
                        }
                        else {
                            this.legendObjectProperties = {};
                        }
                        this.categoryAxisProperties = this.getCategoryAxisProperties(dataViewMetadata);
                        this.valueAxisProperties = this.getValueAxisProperties(dataViewMetadata);
                        var axisPosition = this.valueAxisProperties["position"];
                        this.yAxisOrientation = axisPosition ? axisPosition.toString() : yAxisPosition.left;
                    }
                };
                EnhancedScatterChart.prototype.renderLegend = function () {
                    var legendData = { title: "", dataPoints: [] };
                    var legend = this.legend;
                    this.layerLegendData = this.data.legendData;
                    if (this.layerLegendData) {
                        legendData.title = this.layerLegendData.title || "";
                        legendData.dataPoints = legendData.dataPoints.concat(this.layerLegendData.dataPoints || []);
                        legendData.fontSize = this.legendLabelFontSize ? this.legendLabelFontSize : EnhancedScatterChart.LegendLabelFontSizeDefault;
                        if (this.layerLegendData.grouped) {
                            legendData.grouped = true;
                        }
                    }
                    var legendProperties = this.legendObjectProperties;
                    if (legendProperties) {
                        LegendData.update(legendData, legendProperties);
                        var position = legendProperties[legendProps.position];
                        if (position)
                            legend.changeOrientation(LegendPosition[position]);
                    }
                    else {
                        legend.changeOrientation(LegendPosition.Top);
                    }
                    if (legendData.dataPoints.length === 1 && !legendData.grouped) {
                        legendData.dataPoints = [];
                    }
                    var viewport = this.viewport;
                    legend.drawLegend(legendData, { height: viewport.height, width: viewport.width });
                    Legend.positionChartArea(this.svg, legend);
                };
                EnhancedScatterChart.prototype.shouldRenderAxis = function (axisProperties, propertyName) {
                    if (propertyName === void 0) { propertyName = "show"; }
                    if (!axisProperties) {
                        return false;
                    }
                    else if (axisProperties.isCategoryAxis && (!this.categoryAxisProperties || this.categoryAxisProperties[propertyName] == null || this.categoryAxisProperties[propertyName])) {
                        return axisProperties.values && axisProperties.values.length > 0;
                    }
                    else if (!axisProperties.isCategoryAxis && (!this.valueAxisProperties || this.valueAxisProperties[propertyName] == null || this.valueAxisProperties[propertyName])) {
                        return axisProperties.values && axisProperties.values.length > 0;
                    }
                    return false;
                };
                EnhancedScatterChart.prototype.getMaxMarginFactor = function () {
                    return this.options.style.maxMarginFactor || 0.25;
                };
                EnhancedScatterChart.prototype.adjustViewportbyBackdrop = function () {
                    var img = new Image();
                    var that = this;
                    img.src = this.data.backdrop.url;
                    img.onload = function () {
                        if (that.oldBackdrop !== this.src) {
                            that.render(true);
                            that.oldBackdrop = this.src;
                        }
                    };
                    if (img.width > 0 && img.height > 0) {
                        if (img.width * this.viewportIn.height < this.viewportIn.width * img.height) {
                            var deltaWidth = this.viewportIn.width - this.viewportIn.height * img.width / img.height;
                            this.viewport = { width: this.viewport.width - deltaWidth, height: this.viewport.height };
                        }
                        else {
                            var deltaHeight = this.viewportIn.height - this.viewportIn.width * img.height / img.width;
                            this.viewport = { width: this.viewport.width, height: this.viewport.height - deltaHeight };
                        }
                    }
                };
                EnhancedScatterChart.prototype.render = function (suppressAnimations) {
                    this.viewport.height -= this.legendViewport.height;
                    this.viewport.width -= this.legendViewport.width;
                    if (this.viewportIn.width === 0 || this.viewportIn.height === 0) {
                        return;
                    }
                    var maxMarginFactor = this.getMaxMarginFactor();
                    this.leftRightMarginLimit = this.viewport.width * maxMarginFactor;
                    var bottomMarginLimit = this.bottomMarginLimit = Math.max(25, Math.ceil(this.viewport.height * maxMarginFactor));
                    // reset defaults
                    this.margin.top = 8;
                    this.margin.bottom = bottomMarginLimit;
                    this.margin.right = 0;
                    this.calculateAxes(this.categoryAxisProperties, this.valueAxisProperties, EnhancedScatterChart.TextProperties, true);
                    this.yAxisIsCategorical = this.yAxisProperties.isCategoryAxis;
                    this.hasCategoryAxis = this.yAxisIsCategorical ? this.yAxisProperties && this.yAxisProperties.values.length > 0 : this.xAxisProperties && this.xAxisProperties.values.length > 0;
                    var renderXAxis = this.shouldRenderAxis(this.xAxisProperties);
                    var renderY1Axis = this.shouldRenderAxis(this.yAxisProperties);
                    var mainAxisScale;
                    this.isXScrollBarVisible = false;
                    this.isYScrollBarVisible = false;
                    var tickLabelMargins;
                    var axisLabels;
                    var chartHasAxisLabels;
                    var yAxisOrientation = this.yAxisOrientation;
                    var showY1OnRight = yAxisOrientation === yAxisPosition.right;
                    var axesFontProperties = EnhancedScatterChart.getFontProperties(EnhancedScatterChart.TextProperties.fontFamily, EnhancedScatterChart.AxisFontSize);
                    this.calculateAxes(this.categoryAxisProperties, this.valueAxisProperties, EnhancedScatterChart.TextProperties, true);
                    var doneWithMargins = false, maxIterations = 2, numIterations = 0;
                    while (!doneWithMargins && numIterations < maxIterations) {
                        numIterations++;
                        tickLabelMargins = this.getTickLabelMargins(renderXAxis, renderY1Axis, axesFontProperties, showY1OnRight);
                        // We look at the y axes as main and second sides, if the y axis orientation is right so the main side represents the right side
                        var maxMainYaxisSide = showY1OnRight ? tickLabelMargins.yRight : tickLabelMargins.yLeft, maxSecondYaxisSide = showY1OnRight ? tickLabelMargins.yLeft : tickLabelMargins.yRight, xMax = tickLabelMargins.xMax;
                        maxMainYaxisSide += 10;
                        maxSecondYaxisSide += 10;
                        xMax += 12;
                        if (showY1OnRight && renderY1Axis) {
                            maxSecondYaxisSide += 20;
                        }
                        if (!showY1OnRight && renderY1Axis) {
                            maxMainYaxisSide += 20;
                        }
                        this.addUnitTypeToAxisLabel(this.xAxisProperties, this.yAxisProperties);
                        axisLabels = { x: this.xAxisProperties.axisLabel, y: this.yAxisProperties.axisLabel, y2: null };
                        chartHasAxisLabels = (axisLabels.x != null) || (axisLabels.y != null || axisLabels.y2 != null);
                        if (axisLabels.x != null)
                            xMax += 18;
                        if (axisLabels.y != null)
                            maxMainYaxisSide += 20;
                        if (axisLabels.y2 != null)
                            maxSecondYaxisSide += 20;
                        this.margin.left = showY1OnRight ? maxSecondYaxisSide : maxMainYaxisSide;
                        this.margin.right = showY1OnRight ? maxMainYaxisSide : maxSecondYaxisSide;
                        this.margin.bottom = xMax;
                        // re-calculate the axes with the new margins
                        var previousTickCountY1 = this.yAxisProperties.values.length;
                        this.calculateAxes(this.categoryAxisProperties, this.valueAxisProperties, EnhancedScatterChart.TextProperties, true);
                        // the minor padding adjustments could have affected the chosen tick values, which would then need to calculate margins again
                        // e.g. [0,2,4,6,8] vs. [0,5,10] the 10 is wider and needs more margin.
                        if (this.yAxisProperties.values.length === previousTickCountY1)
                            doneWithMargins = true;
                    }
                    // we have to do the above process again since changes are made to viewport.
                    if (this.data.backdrop && this.data.backdrop.show && (this.data.backdrop.url !== undefined)) {
                        this.adjustViewportbyBackdrop();
                        doneWithMargins = false;
                        maxIterations = 2;
                        numIterations = 0;
                        while (!doneWithMargins && numIterations < maxIterations) {
                            numIterations++;
                            tickLabelMargins = this.getTickLabelMargins(renderXAxis, renderY1Axis, axesFontProperties, showY1OnRight);
                            // We look at the y axes as main and second sides, if the y axis orientation is right so the main side represents the right side
                            var maxMainYaxisSide = showY1OnRight ? tickLabelMargins.yRight : tickLabelMargins.yLeft, maxSecondYaxisSide = showY1OnRight ? tickLabelMargins.yLeft : tickLabelMargins.yRight, xMax = tickLabelMargins.xMax;
                            maxMainYaxisSide += 10;
                            if (showY1OnRight && renderY1Axis) {
                                maxSecondYaxisSide += 15;
                            }
                            xMax += 12;
                            this.addUnitTypeToAxisLabel(this.xAxisProperties, this.yAxisProperties);
                            axisLabels = { x: this.xAxisProperties.axisLabel, y: this.yAxisProperties.axisLabel, y2: null };
                            chartHasAxisLabels = (axisLabels.x != null) || (axisLabels.y != null || axisLabels.y2 != null);
                            if (axisLabels.x != null)
                                xMax += 18;
                            if (axisLabels.y != null)
                                maxMainYaxisSide += 20;
                            if (axisLabels.y2 != null)
                                maxSecondYaxisSide += 20;
                            this.margin.left = showY1OnRight ? maxSecondYaxisSide : maxMainYaxisSide;
                            this.margin.right = showY1OnRight ? maxMainYaxisSide : maxSecondYaxisSide;
                            this.margin.bottom = xMax;
                            // re-calculate the axes with the new margins
                            var previousTickCountY1 = this.yAxisProperties.values.length;
                            this.calculateAxes(this.categoryAxisProperties, this.valueAxisProperties, EnhancedScatterChart.TextProperties, true);
                            // the minor padding adjustments could have affected the chosen tick values, which would then need to calculate margins again
                            // e.g. [0,2,4,6,8] vs. [0,5,10] the 10 is wider and needs more margin.
                            if (this.yAxisProperties.values.length === previousTickCountY1)
                                doneWithMargins = true;
                        }
                    }
                    this.renderChart(mainAxisScale, this.xAxisProperties, this.yAxisProperties, tickLabelMargins, chartHasAxisLabels, axisLabels, suppressAnimations);
                    this.updateAxis();
                    if (!this.data) {
                        return;
                    }
                    var data = this.data, dataPoints = this.data.dataPoints, hasSelection = this.interactivityService && this.interactivityService.hasSelection();
                    this.mainGraphicsSVGSelection
                        .attr("width", this.viewportIn.width)
                        .attr("height", this.viewportIn.height);
                    var sortedData = dataPoints.sort(function (a, b) {
                        return b.radius.sizeMeasure
                            ? (b.radius.sizeMeasure.values[b.radius.index] - a.radius.sizeMeasure.values[a.radius.index])
                            : 0;
                    });
                    var duration = GetAnimationDuration(this.animator, suppressAnimations), scatterMarkers = this.drawScatterMarkers(sortedData, hasSelection, data.sizeRange, duration), dataLabelsSettings = this.data.dataLabelsSettings;
                    if (dataLabelsSettings.show) {
                        var layout, clonedDataPoints, labels;
                        layout = this.getEnhanchedScatterChartLabelLayout(dataLabelsSettings, this.viewportIn, data.sizeRange);
                        clonedDataPoints = this.cloneDataPoints(dataPoints);
                        //fix bug 3863: drawDefaultLabelsForDataPointChart add to datapoints[xxx].size = object , which causes when
                        //category labels is on and Fill Points option off to fill the points when mouse click occures because of default size
                        //is set to datapoints.
                        labels = dataLabelUtils.drawDefaultLabelsForDataPointChart(clonedDataPoints, this.mainGraphicsG, layout, this.viewportIn);
                        if (labels) {
                            labels.attr("transform", function (d) {
                                var size = d.size, dx, dy;
                                dx = size.width / EnhancedScatterChart.DataLabelXOffset;
                                dy = size.height / EnhancedScatterChart.DataLabelYOffset;
                                return SVGUtil.translate(dx, dy);
                            });
                        }
                    }
                    else {
                        dataLabelUtils.cleanDataLabels(this.mainGraphicsG);
                    }
                    this.renderCrosshair(data);
                    var behaviorOptions;
                    if (this.interactivityService) {
                        behaviorOptions = {
                            dataPointsSelection: scatterMarkers,
                            data: this.data,
                            plotContext: this.mainGraphicsSVGSelection,
                        };
                    }
                    addTooltip(scatterMarkers, function (tooltipEvent) { return tooltipEvent.data.tooltipInfo; });
                    SVGUtil.flushAllD3TransitionsIfNeeded(this.options);
                    if (this.behavior) {
                        var layerBehaviorOptions = [behaviorOptions];
                        if (this.interactivityService) {
                            var cbehaviorOptions = {
                                layerOptions: layerBehaviorOptions,
                                clearCatcher: this.clearCatcher,
                            };
                            this.interactivityService.bind(dataPoints, this.behavior, cbehaviorOptions);
                        }
                    }
                };
                EnhancedScatterChart.prototype.getTickLabelMargins = function (renderXAxis, renderY1Axis, axesFontProperties, showY1OnRight) {
                    var labelOptions = this.getTickLabelMarginOptions(renderXAxis, renderY1Axis, axesFontProperties, showY1OnRight);
                    return AxisHelper.getTickLabelMargins(labelOptions, labelOptions.yMarginLimit, labelOptions.textWidthMeasurer, labelOptions.textHeightMeasurer, labelOptions.axes, labelOptions.bottomMarginLimit, EnhancedScatterChart.TextProperties, labelOptions.scrollbarVisible, labelOptions.showOnRight, labelOptions.renderXAxis, labelOptions.renderY1Axis, labelOptions.renderY2Axis);
                };
                EnhancedScatterChart.prototype.getTickLabelMarginOptions = function (renderXAxis, renderY1Axis, axesFontProperties, showY1OnRight) {
                    var width = this.viewportIn.width, height = this.viewport.height;
                    return {
                        width: width,
                        height: height,
                        renderXAxis: renderXAxis,
                        renderY1Axis: renderY1Axis,
                        axesFontProperties: axesFontProperties,
                        axisFontSize: axesFontProperties,
                        plotArea: {
                            width: width,
                            height: height
                        },
                        yMarginLimit: this.leftRightMarginLimit,
                        textWidthMeasurer: measureSvgTextWidth,
                        textHeightMeasurer: measureSvgTextHeight,
                        axes: {
                            x: this.xAxisProperties,
                            y1: this.yAxisProperties
                        },
                        bottomMarginLimit: this.bottomMarginLimit,
                        scrollbarVisible: this.isXScrollBarVisible || this.isYScrollBarVisible,
                        showOnRight: showY1OnRight,
                        renderY2Axis: false
                    };
                };
                /**
                 * This method is temporary in order to fix the visual in PBI Desktop April
                 * We're going to remove it.
                 */
                EnhancedScatterChart.getFontProperties = function (fontFamily, fontSize) {
                    // TODO temprorary solution, remove when not needed
                    var fontProperties = visuals.FontProperties;
                    if (visuals.FontProperties) {
                        if (visuals.FontProperties.applyStyle && FontSize.createFromPx) {
                            return EnhancedScatterChart.getAxesFontProperties(fontFamily, fontSize);
                        }
                        else if (fontProperties.create) {
                            return EnhancedScatterChart.getAxesFontPropertiesOld(fontFamily, fontSize);
                        }
                    }
                    return {
                        x: fontSize,
                        xTitle: fontSize,
                        y: fontSize,
                        yTitle: fontSize,
                        y2: fontSize,
                        y2Title: fontSize
                    };
                };
                // TODO temprorary solution, remove when not needed
                EnhancedScatterChart.getAxesFontPropertiesOld = function (fontFamily, fontSize) {
                    var fontProperties = visuals.FontProperties;
                    return {
                        x: {
                            tickLabels: fontProperties.create({ family: fontFamily, size: { px: fontSize } }),
                            title: fontProperties.create({ family: fontFamily, size: { px: fontSize } })
                        },
                        y: {
                            tickLabels: fontProperties.create({ family: fontFamily, size: { px: fontSize } }),
                            title: fontProperties.create({ family: fontFamily, size: { px: fontSize } })
                        },
                        y2: {
                            tickLabels: fontProperties.create({ family: fontFamily, size: { px: fontSize } }),
                            title: fontProperties.create({ family: fontFamily, size: { px: fontSize } })
                        },
                    };
                };
                EnhancedScatterChart.getAxesFontProperties = function (fontFamily, fontSize) {
                    return {
                        x: {
                            tickLabels: { family: fontFamily, size: FontSize.createFromPx(fontSize) },
                            title: { family: fontFamily, size: FontSize.createFromPx(fontSize) }
                        },
                        y: {
                            tickLabels: { family: fontFamily, size: FontSize.createFromPx(fontSize) },
                            title: { family: fontFamily, size: FontSize.createFromPx(fontSize) }
                        },
                        y2: {
                            tickLabels: { family: fontFamily, size: FontSize.createFromPx(fontSize) },
                            title: { family: fontFamily, size: FontSize.createFromPx(fontSize) }
                        },
                    };
                };
                EnhancedScatterChart.prototype.cloneDataPoints = function (dataPoints) {
                    return dataPoints.map(function (dataPoint) {
                        return _.clone(dataPoint);
                    });
                };
                EnhancedScatterChart.prototype.darkenZeroLine = function (g) {
                    var zeroTick = g.selectAll("g.tick").filter(function (data) { return data === 0; }).node();
                    if (zeroTick) {
                        d3.select(zeroTick).select("line").classed("zero-line", true);
                    }
                };
                EnhancedScatterChart.prototype.getCategoryAxisFill = function () {
                    if (this.dataView && this.dataView.metadata.objects) {
                        var label = this.dataView.metadata.objects["categoryAxis"];
                        if (label) {
                            return label["axisColor"];
                        }
                    }
                    return { solid: { color: "#333" } };
                };
                EnhancedScatterChart.prototype.getEnhanchedScatterChartLabelLayout = function (labelSettings, viewport, sizeRange) {
                    var xScale = this.xAxisProperties.scale;
                    var yScale = this.yAxisProperties.scale;
                    var fontSizeInPx = fromPoint(labelSettings.fontSize);
                    var fontFamily = LabelTextProperties.fontFamily;
                    return {
                        labelText: function (d) {
                            return getLabelFormattedText({
                                label: d.formattedCategory.getValue(),
                                fontSize: labelSettings.fontSize,
                                maxWidth: viewport.width,
                            });
                        },
                        labelLayout: {
                            x: function (d) { return xScale(d.x); },
                            y: function (d) {
                                var margin = EnhancedScatterChart.getBubbleRadius(d.radius, sizeRange, viewport) + dataLabelUtils.labelMargin;
                                return labelSettings.position === 0 /* Above */
                                    ? yScale(d.y) - margin
                                    : yScale(d.y) + margin;
                            },
                        },
                        filter: function (d) { return (d != null && d.formattedCategory.getValue() != null); },
                        style: {
                            "fill": function (d) { return d.labelFill; },
                            "font-size": fontSizeInPx,
                            "font-family": fontFamily,
                        },
                    };
                };
                EnhancedScatterChart.getBubbleRadius = function (radiusData, sizeRange, viewport) {
                    var actualSizeDataRange = null, bubblePixelAreaSizeRange = null, measureSize = radiusData.sizeMeasure;
                    if (!measureSize) {
                        return EnhancedScatterChart.BubbleRadius;
                    }
                    var minSize = sizeRange.min ? sizeRange.min : 0, maxSize = sizeRange.max ? sizeRange.max : 0;
                    var min = Math.min(minSize, 0), max = Math.max(maxSize, 0);
                    actualSizeDataRange = {
                        minRange: min,
                        maxRange: max,
                        delta: max - min
                    };
                    bubblePixelAreaSizeRange = EnhancedScatterChart.getBubblePixelAreaSizeRange(viewport, EnhancedScatterChart.MinSizeRange, EnhancedScatterChart.MaxSizeRange);
                    if (measureSize.values) {
                        var sizeValue = measureSize.values[radiusData.index];
                        if (sizeValue != null) {
                            return EnhancedScatterChart.projectSizeToPixels(sizeValue, actualSizeDataRange, bubblePixelAreaSizeRange) / 2;
                        }
                    }
                    return EnhancedScatterChart.BubbleRadius;
                };
                EnhancedScatterChart.getBubblePixelAreaSizeRange = function (viewPort, minSizeRange, maxSizeRange) {
                    var ratio = 1.0;
                    if (viewPort.height > 0 && viewPort.width > 0) {
                        var minSize = Math.min(viewPort.height, viewPort.width);
                        ratio = (minSize * minSize) / EnhancedScatterChart.AreaOf300By300Chart;
                    }
                    var minRange = Math.round(minSizeRange * ratio), maxRange = Math.round(maxSizeRange * ratio);
                    return {
                        minRange: minRange,
                        maxRange: maxRange,
                        delta: maxRange - minRange
                    };
                };
                EnhancedScatterChart.projectSizeToPixels = function (size, actualSizeDataRange, bubblePixelAreaSizeRange) {
                    var projectedSize = 0;
                    if (actualSizeDataRange) {
                        // Project value on the required range of bubble area sizes
                        projectedSize = bubblePixelAreaSizeRange.maxRange;
                        if (actualSizeDataRange.delta !== 0) {
                            var value = Math.min(Math.max(size, actualSizeDataRange.minRange), actualSizeDataRange.maxRange);
                            projectedSize = EnhancedScatterChart.project(value, actualSizeDataRange, bubblePixelAreaSizeRange);
                        }
                        projectedSize = Math.sqrt(projectedSize / Math.PI) * 2;
                    }
                    return Math.round(projectedSize);
                };
                EnhancedScatterChart.project = function (value, actualSizeDataRange, bubblePixelAreaSizeRange) {
                    if (actualSizeDataRange.delta === 0 || bubblePixelAreaSizeRange.delta === 0) {
                        return (EnhancedScatterChart.rangeContains(actualSizeDataRange, value))
                            ? bubblePixelAreaSizeRange.minRange
                            : null;
                    }
                    var relativeX = (value - actualSizeDataRange.minRange) / actualSizeDataRange.delta;
                    return bubblePixelAreaSizeRange.minRange + relativeX * bubblePixelAreaSizeRange.delta;
                };
                EnhancedScatterChart.rangeContains = function (range, value) {
                    return range.minRange <= value && value <= range.maxRange;
                };
                EnhancedScatterChart.prototype.getValueAxisFill = function () {
                    if (this.dataView && this.dataView.metadata.objects) {
                        var label = this.dataView.metadata.objects["valueAxis"];
                        if (label) {
                            return label["axisColor"];
                        }
                    }
                    return { solid: { color: "#333" } };
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.renderCrosshair = function (data) {
                    if (!this.mainGraphicsSVGSelection) {
                        return;
                    }
                    this.crosshairCanvasSelection = this.addCrosshairCanvasToDOM(this.mainGraphicsSVGSelection);
                    if (data && data.crosshair) {
                        this.crosshairVerticalLineSelection = this.addCrosshairLineToDOM(this.crosshairCanvasSelection, EnhancedScatterChart.CrosshairVerticalLineSelector);
                        this.crosshairHorizontalLineSelection = this.addCrosshairLineToDOM(this.crosshairCanvasSelection, EnhancedScatterChart.CrosshairHorizontalLineSelector);
                        this.crosshairTextSelection = this.addCrosshairTextToDOM(this.crosshairCanvasSelection);
                        this.bindCrosshairEvents();
                    }
                    else {
                        this.clearCrosshair();
                    }
                    return this.crosshairCanvasSelection;
                };
                EnhancedScatterChart.prototype.clearCrosshair = function () {
                    if (!this.crosshairCanvasSelection) {
                        return;
                    }
                    this.crosshairCanvasSelection
                        .selectAll("*")
                        .remove();
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.addCrosshairCanvasToDOM = function (rootElement) {
                    var crosshairCanvasSelector = EnhancedScatterChart.CrosshairCanvasSelector;
                    return this.addElementToDOM(rootElement, {
                        name: "g",
                        selector: crosshairCanvasSelector.selector,
                        className: crosshairCanvasSelector.class,
                        styles: { display: "none" }
                    });
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.addCrosshairLineToDOM = function (rootElement, elementSelector) {
                    var crosshairLineSelector = EnhancedScatterChart.CrosshairLineSelector;
                    return this.addElementToDOM(rootElement, {
                        name: "line",
                        selector: elementSelector.selector,
                        className: crosshairLineSelector.class + " " + elementSelector.class,
                        attributes: { x1: 0, y1: 0, x2: 0, y2: 0 }
                    });
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.addCrosshairTextToDOM = function (rootElement) {
                    var crosshairTextSelector = EnhancedScatterChart.CrosshairTextSelector;
                    return this.addElementToDOM(rootElement, {
                        name: "text",
                        selector: crosshairTextSelector.selector,
                        className: crosshairTextSelector.class
                    });
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.bindCrosshairEvents = function () {
                    var _this = this;
                    if (!this.axisGraphicsContextScrollable) {
                        return;
                    }
                    this.axisGraphicsContextScrollable
                        .on("mousemove", function () {
                        var currentTarget = d3.event.currentTarget, coordinates = d3.mouse(currentTarget), svgNode = currentTarget.viewportElement, scaledRect = svgNode.getBoundingClientRect(), domRect = svgNode.getBBox(), ratioX = scaledRect.width / domRect.width, ratioY = scaledRect.height / domRect.height, x = coordinates[0], y = coordinates[1];
                        if (domRect.width > 0 && !equalWithPrecision(ratioX, 1.0, 0.00001)) {
                            x = x / ratioX;
                        }
                        if (domRect.height > 0 && !equalWithPrecision(ratioY, 1.0, 0.00001)) {
                            y = y / ratioY;
                        }
                        _this.updateCrosshair(x, y);
                    })
                        .on("mouseover", function () {
                        _this.crosshairCanvasSelection.style("display", "block");
                    })
                        .on("mouseout", function () {
                        _this.crosshairCanvasSelection.style("display", "none");
                    });
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.updateCrosshair = function (x, y) {
                    if (!this.viewportIn ||
                        !this.crosshairHorizontalLineSelection ||
                        !this.crosshairVerticalLineSelection ||
                        !this.crosshairTextSelection ||
                        !this.xAxisProperties) {
                        return;
                    }
                    var crosshairTextMargin = EnhancedScatterChart.CrosshairTextMargin, xScale = this.xAxisProperties.scale, yScale = this.yAxisProperties.scale, xFormated, yFormated;
                    this.crosshairHorizontalLineSelection
                        .attr({ x1: 0, y1: y, x2: this.viewportIn.width, y2: y });
                    this.crosshairVerticalLineSelection
                        .attr({ x1: x, y1: 0, x2: x, y2: this.viewportIn.height });
                    xFormated = Math.round(xScale.invert(x) * 100) / 100;
                    yFormated = Math.round(yScale.invert(y) * 100) / 100;
                    this.crosshairTextSelection
                        .attr({ x: x + crosshairTextMargin, y: y - crosshairTextMargin })
                        .text("(" + xFormated + ", " + yFormated + ")");
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.addElementToDOM = function (rootElement, properties) {
                    if (!rootElement || !properties) {
                        return null;
                    }
                    var elementSelection, elementUpdateSelection;
                    elementSelection = rootElement
                        .selectAll(properties.selector);
                    elementUpdateSelection = elementSelection.data(properties.data || [[]]);
                    elementUpdateSelection
                        .enter()
                        .append(properties.name)
                        .attr(properties.attributes)
                        .style(properties.styles)
                        .classed(properties.className, true);
                    elementUpdateSelection
                        .exit()
                        .remove();
                    return elementUpdateSelection;
                };
                EnhancedScatterChart.prototype.renderBackground = function () {
                    if (this.data.backdrop && this.data.backdrop.show && (this.data.backdrop.url !== undefined)) {
                        this.backgroundGraphicsContext
                            .attr("xlink:href", this.data.backdrop.url)
                            .attr("x", 0)
                            .attr("y", 0)
                            .attr("width", this.viewportIn.width)
                            .attr("height", this.viewportIn.height);
                    }
                    else {
                        this.backgroundGraphicsContext
                            .attr("width", 0)
                            .attr("height", 0);
                    }
                };
                EnhancedScatterChart.prototype.renderChart = function (mainAxisScale, xAxis, yAxis, tickLabelMargins, chartHasAxisLabels, axisLabels, suppressAnimations, scrollScale, extent) {
                    var bottomMarginLimit = this.bottomMarginLimit, leftRightMarginLimit = this.leftRightMarginLimit, duration = GetAnimationDuration(this.animator, suppressAnimations);
                    this.renderBackground();
                    //hide show x-axis here
                    if (this.shouldRenderAxis(xAxis)) {
                        xAxis.axis.orient("bottom");
                        if (!xAxis.willLabelsFit)
                            xAxis.axis.tickPadding(5);
                        var xAxisGraphicsElement = this.xAxisGraphicsContext;
                        if (duration) {
                            xAxisGraphicsElement
                                .transition()
                                .duration(duration)
                                .call(xAxis.axis)
                                .call(this.darkenZeroLine);
                        }
                        else {
                            xAxisGraphicsElement
                                .call(xAxis.axis)
                                .call(this.darkenZeroLine);
                        }
                        var xZeroTick = xAxisGraphicsElement
                            .selectAll("g.tick")
                            .filter(function (data) { return data === 0; });
                        if (xZeroTick) {
                            var xZeroColor = this.getValueAxisFill();
                            if (xZeroColor) {
                                xZeroTick
                                    .selectAll("line")
                                    .style({ "stroke": xZeroColor.solid.color });
                            }
                        }
                        var xAxisTextNodes = xAxisGraphicsElement.selectAll("text");
                        if (xAxis.willLabelsWordBreak) {
                            xAxisTextNodes.call(AxisHelper.LabelLayoutStrategy.wordBreak, xAxis, bottomMarginLimit);
                        }
                        else {
                            xAxisTextNodes.call(AxisHelper.LabelLayoutStrategy.rotate, bottomMarginLimit, getTailoredTextOrDefault, EnhancedScatterChart.TextProperties, !xAxis.willLabelsFit, bottomMarginLimit === tickLabelMargins.xMax, xAxis, this.margin, this.isXScrollBarVisible || this.isYScrollBarVisible);
                        }
                    }
                    else {
                        this.xAxisGraphicsContext.selectAll("*").remove();
                    }
                    if (this.shouldRenderAxis(yAxis)) {
                        var yAxisOrientation = this.yAxisOrientation;
                        yAxis.axis
                            .tickSize(-this.viewportIn.width)
                            .tickPadding(10)
                            .orient(yAxisOrientation.toLowerCase());
                        var y1AxisGraphicsElement = this.y1AxisGraphicsContext;
                        if (duration) {
                            y1AxisGraphicsElement
                                .transition()
                                .duration(duration)
                                .call(yAxis.axis)
                                .call(this.darkenZeroLine);
                        }
                        else {
                            y1AxisGraphicsElement
                                .call(yAxis.axis)
                                .call(this.darkenZeroLine);
                        }
                        var yZeroTick = y1AxisGraphicsElement.selectAll("g.tick").filter(function (data) { return data === 0; });
                        if (yZeroTick) {
                            var yZeroColor = this.getCategoryAxisFill();
                            if (yZeroColor) {
                                yZeroTick.selectAll("line").style({ "stroke": yZeroColor.solid.color });
                            }
                        }
                        if (tickLabelMargins.yLeft >= leftRightMarginLimit) {
                            y1AxisGraphicsElement.selectAll("text")
                                .call(AxisHelper.LabelLayoutStrategy.clip, 
                            // Can"t use padding space to render text, so subtract that from available space for ellipses calculations
                            leftRightMarginLimit - 10, svgEllipsis);
                        }
                    }
                    else {
                        this.y1AxisGraphicsContext.selectAll("*").remove();
                    }
                    // Axis labels
                    //TODO: Add label for second Y axis for combo chart
                    if (chartHasAxisLabels) {
                        var hideXAxisTitle = !this.shouldRenderAxis(xAxis, "showAxisTitle");
                        var hideYAxisTitle = !this.shouldRenderAxis(yAxis, "showAxisTitle");
                        var hideY2AxisTitle = this.valueAxisProperties && this.valueAxisProperties["secShowAxisTitle"] != null && this.valueAxisProperties["secShowAxisTitle"] === false;
                        this.renderAxesLabels(axisLabels, this.legendViewport.height, hideXAxisTitle, hideYAxisTitle, hideY2AxisTitle);
                    }
                    else {
                        this.axisGraphicsContext.selectAll(".xAxisLabel").remove();
                        this.axisGraphicsContext.selectAll(".yAxisLabel").remove();
                    }
                };
                EnhancedScatterChart.prototype.renderAxesLabels = function (axisLabels, legendMargin, hideXAxisTitle, hideYAxisTitle, hideY2AxisTitle) {
                    this.axisGraphicsContext.selectAll(".xAxisLabel").remove();
                    this.axisGraphicsContext.selectAll(".yAxisLabel").remove();
                    var margin = this.margin;
                    var width = this.viewportIn.width;
                    var height = this.viewport.height;
                    var fontSize = EnhancedScatterChart.AxisFontSize;
                    var yAxisOrientation = this.yAxisOrientation;
                    var showY1OnRight = yAxisOrientation === yAxisPosition.right;
                    if (!hideXAxisTitle) {
                        var xAxisLabel = this.axisGraphicsContext.append("text")
                            .style("text-anchor", "middle")
                            .text(axisLabels.x)
                            .call(function (text) {
                            text.each(function () {
                                var text = d3.select(this);
                                text.attr({
                                    "class": "xAxisLabel",
                                    "transform": SVGUtil.translate(width / 2, height - fontSize - 2)
                                });
                            });
                        });
                        xAxisLabel.call(AxisHelper.LabelLayoutStrategy.clip, width, svgEllipsis);
                    }
                    if (!hideYAxisTitle) {
                        var yAxisLabel = this.axisGraphicsContext.append("text")
                            .style("text-anchor", "middle")
                            .text(axisLabels.y)
                            .call(function (text) {
                            text.each(function () {
                                var text = d3.select(this);
                                text.attr({
                                    "class": "yAxisLabel",
                                    "transform": "rotate(-90)",
                                    "y": showY1OnRight ? width + margin.right - fontSize : -margin.left,
                                    "x": -((height - margin.top - legendMargin) / 2),
                                    "dy": "1em"
                                });
                            });
                        });
                        yAxisLabel.call(AxisHelper.LabelLayoutStrategy.clip, height - (margin.bottom + margin.top), svgEllipsis);
                    }
                    if (!hideY2AxisTitle && axisLabels.y2) {
                        var y2AxisLabel = this.axisGraphicsContext.append("text")
                            .style("text-anchor", "middle")
                            .text(axisLabels.y2)
                            .call(function (text) {
                            text.each(function () {
                                var text = d3.select(this);
                                text.attr({
                                    "class": "yAxisLabel",
                                    "transform": "rotate(-90)",
                                    "y": showY1OnRight ? -margin.left : width + margin.right - fontSize,
                                    "x": -((height - margin.top - legendMargin) / 2),
                                    "dy": "1em"
                                });
                            });
                        });
                        y2AxisLabel.call(AxisHelper.LabelLayoutStrategy.clip, height - (margin.bottom + margin.top), svgEllipsis);
                    }
                };
                EnhancedScatterChart.prototype.updateAxis = function () {
                    this.adjustMargins();
                    var yAxisOrientation = this.yAxisOrientation;
                    var showY1OnRight = yAxisOrientation === yAxisPosition.right;
                    this.xAxisGraphicsContext
                        .attr("transform", SVGUtil.translate(0, this.viewportIn.height));
                    this.y1AxisGraphicsContext
                        .attr("transform", SVGUtil.translate(showY1OnRight ? this.viewportIn.width : 0, 0));
                    this.svg.attr({
                        "width": this.viewport.width,
                        "height": this.viewport.height
                    });
                    this.svgScrollable.attr({
                        "width": this.viewport.width,
                        "height": this.viewport.height
                    });
                    this.svgScrollable.attr({
                        "x": 0
                    });
                    var left = this.margin.left;
                    var top = this.margin.top;
                    this.axisGraphicsContext.attr("transform", SVGUtil.translate(left, top));
                    this.axisGraphicsContextScrollable.attr("transform", SVGUtil.translate(left, top));
                    this.clearCatcher.attr("transform", SVGUtil.translate(-left, -top));
                    if (this.isXScrollBarVisible) {
                        this.svgScrollable.attr({
                            "x": left
                        });
                        this.axisGraphicsContextScrollable.attr("transform", SVGUtil.translate(0, top));
                        this.svgScrollable.attr("width", this.viewportIn.width);
                        this.svg.attr("width", this.viewport.width)
                            .attr("height", this.viewport.height + this.ScrollBarWidth);
                    }
                    else if (this.isYScrollBarVisible) {
                        this.svgScrollable.attr("height", this.viewportIn.height + top);
                        this.svg.attr("width", this.viewport.width + this.ScrollBarWidth)
                            .attr("height", this.viewport.height);
                    }
                };
                EnhancedScatterChart.prototype.getUnitType = function (xAxis) {
                    if (xAxis.formatter &&
                        xAxis.formatter.displayUnit &&
                        xAxis.formatter.displayUnit.value > 1)
                        return xAxis.formatter.displayUnit.title;
                    return null;
                };
                EnhancedScatterChart.prototype.addUnitTypeToAxisLabel = function (xAxis, yAxis) {
                    var unitType = this.getUnitType(xAxis);
                    if (xAxis.isCategoryAxis) {
                        this.categoryAxisHasUnitType = unitType !== null;
                    }
                    else {
                        this.valueAxisHasUnitType = unitType !== null;
                    }
                    if (xAxis.axisLabel && unitType) {
                        if (xAxis.isCategoryAxis) {
                            xAxis.axisLabel = AxisHelper.createAxisLabel(this.categoryAxisProperties, xAxis.axisLabel, unitType);
                        }
                        else {
                            xAxis.axisLabel = AxisHelper.createAxisLabel(this.valueAxisProperties, xAxis.axisLabel, unitType);
                        }
                    }
                    unitType = this.getUnitType(yAxis);
                    if (!yAxis.isCategoryAxis) {
                        this.valueAxisHasUnitType = unitType !== null;
                    }
                    else {
                        this.categoryAxisHasUnitType = unitType !== null;
                    }
                    if (yAxis.axisLabel && unitType) {
                        if (!yAxis.isCategoryAxis) {
                            yAxis.axisLabel = AxisHelper.createAxisLabel(this.valueAxisProperties, yAxis.axisLabel, unitType);
                        }
                        else {
                            yAxis.axisLabel = AxisHelper.createAxisLabel(this.categoryAxisProperties, yAxis.axisLabel, unitType);
                        }
                    }
                };
                EnhancedScatterChart.prototype.drawScatterMarkers = function (scatterData, hasSelection, sizeRange, duration) {
                    var _this = this;
                    var xScale = this.xAxisProperties.scale, yScale = this.yAxisProperties.scale, shouldEnableFill = (!sizeRange || !sizeRange.min) && this.data.fillPoint;
                    var markers, useCustomColor = this.data.useCustomColor;
                    if (!this.data.useShape) {
                        this.mainGraphicsContext
                            .selectAll(EnhancedScatterChart.ImageClasses.selector)
                            .remove();
                        markers = this.mainGraphicsContext
                            .classed("ScatterMarkers", true)
                            .selectAll(EnhancedScatterChart.DotClasses.selector)
                            .data(scatterData, function (d) { return d.identity.getKey(); });
                        markers
                            .enter()
                            .append("path")
                            .classed(EnhancedScatterChart.DotClasses.class, true)
                            .attr("id", "markershape");
                        markers
                            .style({
                            "stroke-opacity": function (d) {
                                return EnhancedScatterChart.getBubbleOpacity(d, hasSelection);
                            },
                            "stroke-width": "1px",
                            "stroke": function (d) {
                                var color = useCustomColor ? d.colorFill : d.fill;
                                if (_this.data.outline) {
                                    return d3.rgb(color).darker();
                                }
                                else {
                                    return d3.rgb(color);
                                }
                            },
                            "fill": function (d) { return d3.rgb(useCustomColor ? d.colorFill : d.fill); },
                            "fill-opacity": function (d) { return (d.size != null || shouldEnableFill)
                                ? EnhancedScatterChart.getBubbleOpacity(d, hasSelection) : 0; },
                        })
                            .attr("d", function (d) {
                            var r = EnhancedScatterChart.getBubbleRadius(d.radius, sizeRange, _this.viewport), area = 4 * r * r;
                            return d.shapeSymbolType(area);
                        })
                            .transition()
                            .duration(function (d) {
                            if (_this.keyArray.indexOf(d.identity.getKey()) >= 0) {
                                return duration;
                            }
                            else {
                                return 0;
                            }
                        })
                            .attr("transform", function (d) { return "translate(" + xScale(d.x) + "," + yScale(d.y) + ") rotate(" + d.rotation + ")"; });
                    }
                    else {
                        this.mainGraphicsContext
                            .selectAll(EnhancedScatterChart.DotClasses.selector)
                            .remove();
                        markers = this.mainGraphicsContext
                            .classed("ScatterMarkers", true)
                            .selectAll(EnhancedScatterChart.ImageClasses.selector)
                            .data(scatterData, function (d) { return d.identity.getKey(); });
                        markers
                            .enter()
                            .append("svg:image")
                            .classed(EnhancedScatterChart.ImageClasses.class, true)
                            .attr("id", "markerimage");
                        markers
                            .attr("xlink:href", function (d) {
                            if (d.svgurl !== undefined && d.svgurl != null && d.svgurl !== "") {
                                return d.svgurl;
                            }
                            else {
                                return _this.svgDefaultImage;
                            }
                        })
                            .attr("width", function (d) {
                            return EnhancedScatterChart.getBubbleRadius(d.radius, sizeRange, _this.viewport) * 2;
                        })
                            .attr("height", function (d) {
                            return EnhancedScatterChart.getBubbleRadius(d.radius, sizeRange, _this.viewport) * 2;
                        })
                            .transition()
                            .duration(function (d) {
                            if (_this.keyArray.indexOf(d.identity.getKey()) >= 0) {
                                return duration;
                            }
                            else {
                                return 0;
                            }
                        })
                            .attr("transform", function (d) {
                            var radius = EnhancedScatterChart.getBubbleRadius(d.radius, sizeRange, _this.viewport);
                            return "translate(" + (xScale(d.x) - radius) + "," + (yScale(d.y) - radius) + ") rotate(" + d.rotation + "," + radius + "," + radius + ")";
                        });
                    }
                    markers.exit().remove();
                    this.keyArray = [];
                    for (var i = 0; i < scatterData.length; i++) {
                        this.keyArray.push(scatterData[i].identity.getKey());
                    }
                    return markers;
                };
                EnhancedScatterChart.getBubbleOpacity = function (d, hasSelection) {
                    if (hasSelection && !d.selected) {
                        return EnhancedScatterChart.DimmedBubbleOpacity;
                    }
                    return EnhancedScatterChart.DefaultBubbleOpacity;
                };
                EnhancedScatterChart.prototype.calculateAxes = function (categoryAxisProperties, valueAxisProperties, textProperties, scrollbarVisible) {
                    var visualOptions = {
                        viewport: this.viewport,
                        margin: this.margin,
                        forcedXDomain: [categoryAxisProperties ? categoryAxisProperties["start"] : null, categoryAxisProperties ? categoryAxisProperties["end"] : null],
                        forceMerge: valueAxisProperties && valueAxisProperties["secShow"] === false,
                        categoryAxisScaleType: categoryAxisProperties && categoryAxisProperties["axisScale"] != null ? categoryAxisProperties["axisScale"] : null,
                        valueAxisScaleType: valueAxisProperties && valueAxisProperties["axisScale"] != null ? valueAxisProperties["axisScale"] : null,
                        valueAxisDisplayUnits: valueAxisProperties && valueAxisProperties["labelDisplayUnits"] != null ? valueAxisProperties["labelDisplayUnits"] : EnhancedScatterChart.LabelDisplayUnitsDefault,
                        categoryAxisDisplayUnits: categoryAxisProperties && categoryAxisProperties["labelDisplayUnits"] != null ? categoryAxisProperties["labelDisplayUnits"] : EnhancedScatterChart.LabelDisplayUnitsDefault,
                        trimOrdinalDataOnOverflow: false
                    };
                    if (valueAxisProperties) {
                        visualOptions.forcedYDomain = AxisHelper.applyCustomizedDomain([valueAxisProperties["start"], valueAxisProperties["end"]], visualOptions.forcedYDomain);
                    }
                    var width = this.viewport.width - (this.margin.left + this.margin.right);
                    var axes = this.calculateAxesProperties(visualOptions);
                    axes[0].willLabelsFit = AxisHelper.LabelLayoutStrategy.willLabelsFit(axes[0], width, measureSvgTextWidth, textProperties);
                    // If labels do not fit and we are not scrolling, try word breaking
                    axes[0].willLabelsWordBreak = (!axes[0].willLabelsFit && !scrollbarVisible) && AxisHelper.LabelLayoutStrategy.willLabelsWordBreak(axes[0], this.margin, width, measureSvgTextWidth, estimateSvgTextHeight, getTailoredTextOrDefault, textProperties);
                    return axes;
                };
                EnhancedScatterChart.prototype.calculateAxesProperties = function (options) {
                    var data = this.data, dataPoints = data.dataPoints;
                    this.margin = options.margin;
                    this.viewport = options.viewport;
                    var minY = 0, maxY = 10, minX = 0, maxX = 10;
                    if (dataPoints.length > 0) {
                        minY = d3.min(dataPoints, function (d) { return d.y; });
                        maxY = d3.max(dataPoints, function (d) { return d.y; });
                        minX = d3.min(dataPoints, function (d) { return d.x; });
                        maxX = d3.max(dataPoints, function (d) { return d.x; });
                    }
                    var xDomain = [minX, maxX], combinedXDomain, combinedYDomain, xAxisFormatString, yAxisFormatString;
                    combinedXDomain = AxisHelper.combineDomain(this.optimizeTranslateValues(options.forcedXDomain), xDomain);
                    xAxisFormatString = valueFormatter.getFormatString(data.xCol, EnhancedScatterChart.Properties["general"]["formatString"]);
                    this.xAxisProperties = AxisHelper.createAxis({
                        pixelSpan: this.viewportIn.width,
                        dataDomain: combinedXDomain,
                        metaDataColumn: data.xCol,
                        formatString: xAxisFormatString,
                        outerPadding: 0,
                        isScalar: true,
                        isVertical: false,
                        forcedTickCount: options.forcedTickCount,
                        useTickIntervalForDisplayUnits: true,
                        isCategoryAxis: true,
                        scaleType: options.categoryAxisScaleType,
                        axisDisplayUnits: options.categoryAxisDisplayUnits,
                        margin: this.margin,
                    });
                    this.xAxisProperties.axis.tickSize(-this.viewportIn.height, 0);
                    this.xAxisProperties.axisLabel = this.data.axesLabels.x;
                    combinedYDomain = AxisHelper.combineDomain(this.optimizeTranslateValues(options.forcedYDomain), [minY, maxY]);
                    yAxisFormatString = valueFormatter.getFormatString(data.yCol, EnhancedScatterChart.Properties["general"]["formatString"]);
                    this.yAxisProperties = AxisHelper.createAxis({
                        pixelSpan: this.viewportIn.height,
                        dataDomain: combinedYDomain,
                        metaDataColumn: data.yCol,
                        formatString: yAxisFormatString,
                        outerPadding: 0,
                        isScalar: true,
                        isVertical: true,
                        forcedTickCount: options.forcedTickCount,
                        useTickIntervalForDisplayUnits: true,
                        isCategoryAxis: false,
                        scaleType: options.valueAxisScaleType,
                        axisDisplayUnits: options.valueAxisDisplayUnits
                    });
                    this.yAxisProperties.axisLabel = this.data.axesLabels.y;
                    return [this.xAxisProperties, this.yAxisProperties];
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.optimizeTranslateValues = function (values) {
                    var _this = this;
                    if (values && values.map) {
                        return values.map(function (value) {
                            return _this.optimizeTranslateValue(value);
                        });
                    }
                    return values;
                };
                /**
                 * Public for testability.
                 */
                EnhancedScatterChart.prototype.optimizeTranslateValue = function (value) {
                    if (value) {
                        var numberSign = value >= 0 ? 1 : -1, absoluteValue = Math.abs(value);
                        if (absoluteValue > EnhancedScatterChart.MaxTranslateValue) {
                            return EnhancedScatterChart.MaxTranslateValue * numberSign;
                        }
                        else if (absoluteValue < EnhancedScatterChart.MinTranslateValue) {
                            return EnhancedScatterChart.MinTranslateValue * numberSign;
                        }
                    }
                    return value;
                };
                EnhancedScatterChart.prototype.enumerateDataPoints = function (enumeration) {
                    var data = this.data;
                    if (!data)
                        return;
                    var seriesCount = data.dataPoints.length;
                    if (!data.hasDynamicSeries) {
                        var showAllDataPoints = data.showAllDataPoints;
                        // Add default color and show all slices
                        enumeration.pushInstance({
                            objectName: "dataPoint",
                            selector: null,
                            properties: {
                                defaultColor: {
                                    solid: { color: data.defaultDataPointColor || this.colors.getColorByIndex(0).value }
                                }
                            }
                        }).pushInstance({
                            objectName: "dataPoint",
                            selector: null,
                            properties: { showAllDataPoints: showAllDataPoints }
                        });
                        if (showAllDataPoints) {
                            for (var i = 0; i < seriesCount; i++) {
                                var seriesDataPoints = data.dataPoints[i];
                                enumeration.pushInstance({
                                    objectName: "dataPoint",
                                    displayName: seriesDataPoints.formattedCategory.getValue(),
                                    selector: ColorHelper.normalizeSelector(seriesDataPoints.identity.getSelector(), /*isSingleSeries*/ true),
                                    properties: {
                                        fill: { solid: { color: seriesDataPoints.fill } }
                                    },
                                });
                            }
                        }
                    }
                    else {
                        var legendDataPointLength = data.legendData.dataPoints.length;
                        for (var i = 0; i < legendDataPointLength; i++) {
                            var series = data.legendData.dataPoints[i];
                            enumeration.pushInstance({
                                objectName: "dataPoint",
                                displayName: series.label,
                                selector: ColorHelper.normalizeSelector(series.identity.getSelector()),
                                properties: {
                                    fill: { solid: { color: series.color } }
                                },
                            });
                        }
                    }
                };
                EnhancedScatterChart.prototype.enumerateObjectInstances = function (options) {
                    var enumeration = new ObjectEnumerationBuilder();
                    switch (options.objectName) {
                        case "dataPoint": {
                            var categoricalDataView = this.dataView && this.dataView.categorical
                                ? this.dataView.categorical
                                : null;
                            if (!GradientUtils.hasGradientRole(categoricalDataView)) {
                                this.enumerateDataPoints(enumeration);
                            }
                            break;
                        }
                        case "categoryAxis": {
                            this.getCategoryAxisValues(enumeration);
                            break;
                        }
                        case "valueAxis": {
                            this.getValueAxisValues(enumeration);
                            break;
                        }
                        case "categoryLabels": {
                            if (this.data) {
                                dataLabelUtils.enumerateCategoryLabels(enumeration, this.data.dataLabelsSettings, true);
                            }
                            else {
                                dataLabelUtils.enumerateCategoryLabels(enumeration, null, true);
                            }
                            break;
                        }
                        case "fillPoint": {
                            var sizeRange = this.data.sizeRange;
                            // Check if the card should be shown or not
                            if (sizeRange && sizeRange.min) {
                                break;
                            }
                            enumeration.pushInstance({
                                objectName: "fillPoint",
                                selector: null,
                                properties: {
                                    show: this.data.fillPoint,
                                },
                            });
                            break;
                        }
                        case "backdrop": {
                            enumeration.pushInstance({
                                objectName: "backdrop",
                                displayName: "Backdrop",
                                selector: null,
                                properties: {
                                    show: this.data.backdrop ? this.data.backdrop.show : false,
                                    url: this.data.backdrop ? this.data.backdrop.url : null
                                },
                            });
                            break;
                        }
                        case "crosshair": {
                            enumeration.pushInstance({
                                objectName: "crosshair",
                                selector: null,
                                properties: {
                                    show: this.data.crosshair
                                },
                            });
                            break;
                        }
                        case "outline": {
                            enumeration.pushInstance({
                                objectName: "outline",
                                selector: null,
                                properties: {
                                    show: this.data.outline
                                },
                            });
                            break;
                        }
                        case "legend": {
                            this.enumerateLegend(enumeration);
                            break;
                        }
                    }
                    return enumeration.complete() || [];
                };
                EnhancedScatterChart.prototype.hasLegend = function () {
                    return this.data && this.data.hasDynamicSeries;
                };
                EnhancedScatterChart.prototype.enumerateLegend = function (enumeration) {
                    if (!this.hasLegend()) {
                        return;
                    }
                    var show = DataViewObject.getValue(this.legendObjectProperties, legendProps.show, this.legend.isVisible());
                    var showTitle = DataViewObject.getValue(this.legendObjectProperties, legendProps.showTitle, true);
                    var titleText = DataViewObject.getValue(this.legendObjectProperties, legendProps.titleText, this.layerLegendData ? this.layerLegendData.title : "");
                    var legendLabelColor = DataViewObject.getValue(this.legendObjectProperties, legendProps.labelColor, LegendData.DefaultLegendLabelFillColor);
                    this.legendLabelFontSize = DataViewObject.getValue(this.legendObjectProperties, legendProps.fontSize, EnhancedScatterChart.LegendLabelFontSizeDefault);
                    var position = DataViewObject.getValue(this.legendObjectProperties, legendProps.position, legendPosition.top);
                    enumeration.pushInstance({
                        selector: null,
                        properties: {
                            show: show,
                            position: position,
                            showTitle: showTitle,
                            titleText: titleText,
                            labelColor: legendLabelColor,
                            fontSize: this.legendLabelFontSize,
                        },
                        objectName: "legend"
                    });
                };
                EnhancedScatterChart.prototype.getCategoryAxisValues = function (enumeration) {
                    var supportedType = axisType.both, isScalar = true, logPossible = false, scaleOptions = [
                        axisScale.log,
                        axisScale.linear
                    ]; //until options can be update in propPane, show all options
                    if (!isScalar) {
                        if (this.categoryAxisProperties) {
                            this.categoryAxisProperties["start"] = null;
                            this.categoryAxisProperties["end"] = null;
                        }
                    }
                    var instance = {
                        selector: null,
                        properties: {},
                        objectName: "categoryAxis",
                        validValues: {
                            axisScale: scaleOptions
                        }
                    };
                    instance.properties["show"] = this.categoryAxisProperties && this.categoryAxisProperties["show"] != null ? this.categoryAxisProperties["show"] : true;
                    if (this.yAxisIsCategorical)
                        instance.properties["position"] = this.valueAxisProperties && this.valueAxisProperties["position"] != null ? this.valueAxisProperties["position"] : yAxisPosition.left;
                    if (supportedType === axisType.both) {
                        instance.properties["axisType"] = isScalar ? axisType.scalar : axisType.categorical;
                    }
                    if (isScalar) {
                        instance.properties["axisScale"] = (this.categoryAxisProperties && this.categoryAxisProperties["axisScale"] != null && logPossible) ? this.categoryAxisProperties["axisScale"] : axisScale.linear;
                        instance.properties["start"] = this.categoryAxisProperties ? this.categoryAxisProperties["start"] : null;
                        instance.properties["end"] = this.categoryAxisProperties ? this.categoryAxisProperties["end"] : null;
                        instance.properties["labelDisplayUnits"] = this.categoryAxisProperties && this.categoryAxisProperties["labelDisplayUnits"] != null ? this.categoryAxisProperties["labelDisplayUnits"] : EnhancedScatterChart.LabelDisplayUnitsDefault;
                    }
                    instance.properties["showAxisTitle"] = this.categoryAxisProperties && this.categoryAxisProperties["showAxisTitle"] != null ? this.categoryAxisProperties["showAxisTitle"] : true;
                    enumeration
                        .pushInstance(instance)
                        .pushInstance({
                        selector: null,
                        properties: {
                            axisStyle: this.categoryAxisProperties && this.categoryAxisProperties["axisStyle"]
                                ? this.categoryAxisProperties["axisStyle"] : axisStyle.showTitleOnly,
                            labelColor: this.categoryAxisProperties ? this.categoryAxisProperties["labelColor"] : null
                        },
                        objectName: "categoryAxis",
                        validValues: {
                            axisStyle: this.categoryAxisHasUnitType
                                ? [axisStyle.showTitleOnly, axisStyle.showUnitOnly, axisStyle.showBoth]
                                : [axisStyle.showTitleOnly]
                        }
                    });
                };
                //todo: wrap all these object getters and other related stuff into an interface
                EnhancedScatterChart.prototype.getValueAxisValues = function (enumeration) {
                    var scaleOptions = [axisScale.log, axisScale.linear]; //until options can be update in propPane, show all options
                    var logPossible = false;
                    var instance = {
                        selector: null,
                        properties: {},
                        objectName: "valueAxis",
                        validValues: {
                            axisScale: scaleOptions,
                            secAxisScale: scaleOptions
                        }
                    };
                    instance.properties["show"] = this.valueAxisProperties && this.valueAxisProperties["show"] != null ? this.valueAxisProperties["show"] : true;
                    if (!this.yAxisIsCategorical) {
                        instance.properties["position"] = this.valueAxisProperties && this.valueAxisProperties["position"] != null ? this.valueAxisProperties["position"] : yAxisPosition.left;
                    }
                    instance.properties["axisScale"] = (this.valueAxisProperties && this.valueAxisProperties["axisScale"] != null && logPossible) ? this.valueAxisProperties["axisScale"] : axisScale.linear;
                    instance.properties["start"] = this.valueAxisProperties ? this.valueAxisProperties["start"] : null;
                    instance.properties["end"] = this.valueAxisProperties ? this.valueAxisProperties["end"] : null;
                    instance.properties["showAxisTitle"] = this.valueAxisProperties && this.valueAxisProperties["showAxisTitle"] != null ? this.valueAxisProperties["showAxisTitle"] : true;
                    instance.properties["labelDisplayUnits"] = this.valueAxisProperties && this.valueAxisProperties["labelDisplayUnits"] != null ? this.valueAxisProperties["labelDisplayUnits"] : EnhancedScatterChart.LabelDisplayUnitsDefault;
                    enumeration
                        .pushInstance(instance)
                        .pushInstance({
                        selector: null,
                        properties: {
                            axisStyle: this.valueAxisProperties && this.valueAxisProperties["axisStyle"] != null ? this.valueAxisProperties["axisStyle"] : axisStyle.showTitleOnly,
                            labelColor: this.valueAxisProperties ? this.valueAxisProperties["labelColor"] : null
                        },
                        objectName: "valueAxis",
                        validValues: {
                            axisStyle: this.valueAxisHasUnitType ? [axisStyle.showTitleOnly, axisStyle.showUnitOnly, axisStyle.showBoth] : [axisStyle.showTitleOnly]
                        },
                    });
                };
                EnhancedScatterChart.prototype.onClearSelection = function () {
                    if (this.interactivityService)
                        this.interactivityService.clearSelection();
                };
                return EnhancedScatterChart;
            }());
            EnhancedScatterChart.AxisGraphicsContextClassName = "axisGraphicsContext";
            EnhancedScatterChart.ClassName = "enhancedScatterChart";
            EnhancedScatterChart.MainGraphicsContextClassName = "mainGraphicsContext";
            EnhancedScatterChart.LegendLabelFontSizeDefault = 9;
            EnhancedScatterChart.LabelDisplayUnitsDefault = 0;
            EnhancedScatterChart.AxisFontSize = 11;
            EnhancedScatterChart.CrosshairTextMargin = 5;
            EnhancedScatterChart.BubbleRadius = 3 * 2;
            EnhancedScatterChart.MinSizeRange = 200;
            EnhancedScatterChart.MaxSizeRange = 3000;
            EnhancedScatterChart.AreaOf300By300Chart = 90000;
            EnhancedScatterChart.DataLabelXOffset = 2;
            EnhancedScatterChart.DataLabelYOffset = 1.8;
            EnhancedScatterChart.DotClasses = createClassAndSelector("dot");
            EnhancedScatterChart.ImageClasses = createClassAndSelector("img");
            EnhancedScatterChart.TextProperties = {
                fontFamily: "'Segoe UI', wf_segoe-ui_normal, helvetica, arial, sans-serif",
                fontSize: PixelConverter.toString(EnhancedScatterChart.AxisFontSize),
            };
            EnhancedScatterChart.CrosshairCanvasSelector = createClassAndSelector("crosshairCanvas");
            EnhancedScatterChart.CrosshairLineSelector = createClassAndSelector("crosshairLine");
            EnhancedScatterChart.CrosshairVerticalLineSelector = createClassAndSelector("crosshairVerticalLine");
            EnhancedScatterChart.CrosshairHorizontalLineSelector = createClassAndSelector("crosshairHorizontalLine");
            EnhancedScatterChart.CrosshairTextSelector = createClassAndSelector("crosshairText");
            EnhancedScatterChart.MaxTranslateValue = 1e+25;
            EnhancedScatterChart.MinTranslateValue = 1e-25;
            EnhancedScatterChart.DefaultBubbleOpacity = 0.85;
            EnhancedScatterChart.DimmedBubbleOpacity = 0.4;
            EnhancedScatterChart.ColumnCategory = "Category";
            EnhancedScatterChart.ColumnSeries = "Series";
            EnhancedScatterChart.ColumnX = "X";
            EnhancedScatterChart.ColumnY = "Y";
            EnhancedScatterChart.ColumnSize = "Size";
            EnhancedScatterChart.ColumnGradient = "Gradient";
            EnhancedScatterChart.ColumnColorFill = "ColorFill";
            EnhancedScatterChart.ColumnShape = "Shape";
            EnhancedScatterChart.ColumnImage = "Image";
            EnhancedScatterChart.ColumnRotation = "Rotation";
            EnhancedScatterChart.ColumnBackdrop = "Backdrop";
            EnhancedScatterChart.ColumnXStart = "X Start";
            EnhancedScatterChart.ColumnXEnd = "X End";
            EnhancedScatterChart.ColumnYStart = "Y Start";
            EnhancedScatterChart.ColumnYEnd = "Y End";
            EnhancedScatterChart.capabilities = {
                dataRoles: [
                    {
                        name: EnhancedScatterChart.ColumnCategory,
                        kind: VisualDataRoleKind.Grouping,
                        displayName: "Details",
                    }, {
                        name: EnhancedScatterChart.ColumnSeries,
                        kind: VisualDataRoleKind.Grouping,
                        displayName: "Legend",
                    }, {
                        name: EnhancedScatterChart.ColumnX,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "X Axis",
                    }, {
                        name: EnhancedScatterChart.ColumnY,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "Y Axis",
                    }, {
                        name: EnhancedScatterChart.ColumnSize,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "Size",
                    }, {
                        name: EnhancedScatterChart.ColumnGradient,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "Color saturation",
                    }, {
                        name: EnhancedScatterChart.ColumnColorFill,
                        kind: VisualDataRoleKind.Grouping,
                        displayName: "Customized Color",
                    }, {
                        name: EnhancedScatterChart.ColumnShape,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "Shape",
                    }, {
                        name: EnhancedScatterChart.ColumnImage,
                        kind: VisualDataRoleKind.Grouping,
                        displayName: "Image",
                    }, {
                        name: EnhancedScatterChart.ColumnRotation,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "Rotation",
                    }, {
                        name: EnhancedScatterChart.ColumnBackdrop,
                        kind: VisualDataRoleKind.Grouping,
                        displayName: "Backdrop",
                    }, {
                        name: EnhancedScatterChart.ColumnXStart,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "X Start",
                    }, {
                        name: EnhancedScatterChart.ColumnXEnd,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "X End",
                    }, {
                        name: EnhancedScatterChart.ColumnYStart,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "Y Start",
                    }, {
                        name: EnhancedScatterChart.ColumnYEnd,
                        kind: VisualDataRoleKind.Measure,
                        displayName: "Y End",
                    }
                ],
                dataViewMappings: [{
                        conditions: [
                            {
                                "Category": { max: 1 },
                                "Series": { max: 0 },
                                "X": { max: 1 },
                                "Y": { max: 1 },
                                "Size": { max: 1 },
                                "Gradient": { max: 1 },
                                "ColorFill": { max: 0 },
                                "Shape": { max: 1 },
                                "Image": { max: 0 },
                                "Rotation": { max: 1 },
                                "Backdrop": { max: 1 },
                                "X Start": { max: 1 },
                                "X End": { max: 1 },
                                "Y Start": { max: 1 },
                                "Y End": { max: 1 }
                            }, {
                                "Category": { max: 1 },
                                "Series": { max: 0 },
                                "X": { max: 1 },
                                "Y": { max: 1 },
                                "Size": { max: 1 },
                                "Gradient": { max: 0 },
                                "ColorFill": { max: 1 },
                                "Shape": { max: 1 },
                                "Image": { max: 0 },
                                "Rotation": { max: 1 },
                                "Backdrop": { max: 1 },
                                "X Start": { max: 1 },
                                "X End": { max: 1 },
                                "Y Start": { max: 1 },
                                "Y End": { max: 1 }
                            }, {
                                "Category": { max: 1 },
                                "Series": { max: 0 },
                                "X": { max: 1 },
                                "Y": { max: 1 },
                                "Size": { max: 1 },
                                "Gradient": { max: 0 },
                                "ColorFill": { max: 0 },
                                "Shape": { max: 0 },
                                "Image": { max: 1 },
                                "Rotation": { max: 1 },
                                "Backdrop": { max: 1 },
                                "X Start": { max: 1 },
                                "X End": { max: 1 },
                                "Y Start": { max: 1 },
                                "Y End": { max: 1 }
                            }, {
                                "Category": { max: 1 },
                                "Series": { max: 1 },
                                "X": { max: 1 },
                                "Y": { max: 1 },
                                "Size": { max: 1 },
                                "Gradient": { max: 0 },
                                "ColorFill": { max: 1 },
                                "Shape": { max: 1 },
                                "Image": { max: 0 },
                                "Rotation": { max: 1 },
                                "Backdrop": { max: 1 },
                                "X Start": { max: 1 },
                                "X End": { max: 1 },
                                "Y Start": { max: 1 },
                                "Y End": { max: 1 }
                            }, {
                                "Category": { max: 1 },
                                "Series": { max: 1 },
                                "X": { max: 1 },
                                "Y": { max: 1 },
                                "Size": { max: 1 },
                                "Gradient": { max: 0 },
                                "ColorFill": { max: 0 },
                                "Shape": { max: 0 },
                                "Image": { max: 1 },
                                "Rotation": { max: 1 },
                                "Backdrop": { max: 1 },
                                "X Start": { max: 1 },
                                "X End": { max: 1 },
                                "Y Start": { max: 1 },
                                "Y End": { max: 1 }
                            }
                        ],
                        categorical: {
                            categories: {
                                select: [
                                    { bind: { to: EnhancedScatterChart.ColumnCategory } },
                                    { bind: { to: EnhancedScatterChart.ColumnColorFill } },
                                    { bind: { to: EnhancedScatterChart.ColumnImage } },
                                    { bind: { to: EnhancedScatterChart.ColumnBackdrop } }
                                ],
                                dataReductionAlgorithm: { sample: {} }
                            },
                            values: {
                                group: {
                                    by: EnhancedScatterChart.ColumnSeries,
                                    select: [
                                        { bind: { to: EnhancedScatterChart.ColumnX } },
                                        { bind: { to: EnhancedScatterChart.ColumnY } },
                                        { bind: { to: EnhancedScatterChart.ColumnSize } },
                                        { bind: { to: EnhancedScatterChart.ColumnGradient } },
                                        { bind: { to: EnhancedScatterChart.ColumnShape } },
                                        { bind: { to: EnhancedScatterChart.ColumnRotation } },
                                        { bind: { to: EnhancedScatterChart.ColumnXStart } },
                                        { bind: { to: EnhancedScatterChart.ColumnXEnd } },
                                        { bind: { to: EnhancedScatterChart.ColumnYStart } },
                                        { bind: { to: EnhancedScatterChart.ColumnYEnd } }
                                    ],
                                    dataReductionAlgorithm: { top: {} }
                                }
                            },
                            rowCount: { preferred: { min: 2 } }
                        },
                    }],
                objects: {
                    dataPoint: {
                        displayName: "Data colors",
                        properties: {
                            defaultColor: {
                                displayName: "Default color",
                                type: { fill: { solid: { color: true } } }
                            },
                            showAllDataPoints: {
                                displayName: "Show all",
                                type: { bool: true }
                            },
                            fill: {
                                displayName: "Fill",
                                type: { fill: { solid: { color: true } } }
                            },
                            fillRule: {
                                displayName: "Color saturation",
                                type: { fillRule: {} },
                                rule: {
                                    inputRole: EnhancedScatterChart.ColumnGradient,
                                    output: {
                                        property: "fill",
                                        selector: [EnhancedScatterChart.ColumnCategory],
                                    },
                                }
                            }
                        }
                    },
                    general: {
                        displayName: "General",
                        properties: {
                            formatString: {
                                type: { formatting: { formatString: true } },
                            },
                        },
                    },
                    categoryAxis: {
                        displayName: "X-Axis",
                        properties: {
                            show: {
                                displayName: "Show",
                                type: { bool: true }
                            },
                            axisScale: {
                                displayName: "Scale type",
                                type: { formatting: { axisScale: true } }
                            },
                            start: {
                                displayName: "Start",
                                type: { numeric: true }
                            },
                            end: {
                                displayName: "End",
                                type: { numeric: true }
                            },
                            showAxisTitle: {
                                displayName: "Title",
                                type: { bool: true }
                            },
                            axisStyle: {
                                displayName: "Style",
                                type: { formatting: { axisStyle: true } }
                            },
                            axisColor: {
                                displayName: "Color",
                                type: { fill: { solid: { color: true } } }
                            },
                            labelDisplayUnits: {
                                displayName: "Display Units",
                                type: { formatting: { labelDisplayUnits: true } },
                            },
                        }
                    },
                    valueAxis: {
                        displayName: "Y-Axis",
                        properties: {
                            show: {
                                displayName: "Show",
                                type: { bool: true }
                            },
                            position: {
                                displayName: "Position",
                                type: { formatting: { yAxisPosition: true } }
                            },
                            axisScale: {
                                displayName: "Scale type",
                                type: { formatting: { axisScale: true } }
                            },
                            start: {
                                displayName: "Start",
                                type: { numeric: true }
                            },
                            end: {
                                displayName: "End",
                                type: { numeric: true }
                            },
                            showAxisTitle: {
                                displayName: "Title",
                                type: { bool: true }
                            },
                            axisStyle: {
                                displayName: "Style",
                                type: { formatting: { axisStyle: true } }
                            },
                            axisColor: {
                                displayName: "Color",
                                type: { fill: { solid: { color: true } } }
                            },
                            labelDisplayUnits: {
                                displayName: "Display Units",
                                type: { formatting: { labelDisplayUnits: true } },
                            }
                        }
                    },
                    legend: {
                        displayName: "Legend",
                        properties: {
                            show: {
                                displayName: "Show",
                                type: { bool: true }
                            },
                            position: {
                                displayName: "Position",
                                description: "Select the location for the legend",
                                type: { enumeration: legendPosition.type },
                            },
                            showTitle: {
                                displayName: "Title",
                                description: "Display a title for legend symbols",
                                type: { bool: true }
                            },
                            titleText: {
                                displayName: "Legend Name",
                                description: "Title text",
                                type: { text: true }
                            },
                            labelColor: {
                                displayName: "Color",
                                type: { fill: { solid: { color: true } } }
                            },
                            fontSize: {
                                displayName: "Text Size",
                                type: { formatting: { fontSize: true } }
                            }
                        }
                    },
                    categoryLabels: {
                        displayName: "Category labels",
                        properties: {
                            show: {
                                displayName: "Show",
                                type: { bool: true }
                            },
                            color: {
                                displayName: "Color",
                                type: { fill: { solid: { color: true } } }
                            },
                            fontSize: {
                                displayName: "Text Size",
                                type: { formatting: { fontSize: true } }
                            },
                        },
                    },
                    fillPoint: {
                        displayName: "Fill point",
                        properties: {
                            show: {
                                displayName: "Fill",
                                type: { bool: true }
                            },
                        },
                    },
                    backdrop: {
                        displayName: "Backdrop",
                        properties: {
                            show: {
                                displayName: "Show",
                                type: { bool: true }
                            },
                            url: {
                                displayName: "Image URL",
                                type: { text: true }
                            },
                        },
                    },
                    crosshair: {
                        displayName: "Crosshair",
                        properties: {
                            show: {
                                displayName: "Crosshair",
                                type: { bool: true }
                            },
                        },
                    },
                    outline: {
                        displayName: "Outline",
                        properties: {
                            show: {
                                displayName: "Outline",
                                type: { bool: true }
                            }
                        }
                    }
                }
            };
            /**
             * Public for testability.
             */
            EnhancedScatterChart.Properties = EnhancedScatterChart.getPropertiesByCapabilities(EnhancedScatterChart.capabilities);
            EnhancedScatterChart1443994985041.EnhancedScatterChart = EnhancedScatterChart;
            var CustomVisualBehavior = (function () {
                function CustomVisualBehavior(behaviors) {
                    this.behaviors = behaviors || [];
                }
                CustomVisualBehavior.prototype.bindEvents = function (options, selectionHandler) {
                    var behaviors = this.behaviors;
                    for (var i = 0, ilen = behaviors.length; i < ilen; i++) {
                        behaviors[i].bindEvents(options.layerOptions[i], selectionHandler);
                    }
                    options.clearCatcher.on("click", function () {
                        selectionHandler.handleClearSelection();
                    });
                };
                CustomVisualBehavior.prototype.renderSelection = function (hasSelection) {
                    for (var _i = 0, _a = this.behaviors; _i < _a.length; _i++) {
                        var behavior = _a[_i];
                        behavior.renderSelection(hasSelection);
                    }
                };
                return CustomVisualBehavior;
            }());
            EnhancedScatterChart1443994985041.CustomVisualBehavior = CustomVisualBehavior;
            var EnhancedScatterChartWebBehavior = (function () {
                function EnhancedScatterChartWebBehavior(dimmedBubbleOpacity, defaultBubbleOpacity) {
                    this.dimmedBubbleOpacity = dimmedBubbleOpacity;
                    this.defaultBubbleOpacity = defaultBubbleOpacity;
                }
                EnhancedScatterChartWebBehavior.prototype.bindEvents = function (options, selectionHandler) {
                    var bubbles = this.bubbles = options.dataPointsSelection, data = options.data;
                    this.shouldEnableFill = (!data.sizeRange || !data.sizeRange.min) && data.fillPoint;
                    this.colorBorder = data.colorBorder;
                    registerStandardInteractivityHandlers(bubbles, selectionHandler);
                };
                EnhancedScatterChartWebBehavior.prototype.renderSelection = function (hasSelection) {
                    var _this = this;
                    var shouldEnableFill = this.shouldEnableFill, colorBorder = this.colorBorder;
                    this.bubbles.style("fill-opacity", function (d) {
                        return _this.getMarkerFillOpacity(d.size != null, shouldEnableFill, hasSelection, d.selected);
                    });
                    this.bubbles.style("stroke-opacity", function (d) {
                        return _this.getMarkerStrokeOpacity(d.size != null, colorBorder, hasSelection, d.selected);
                    });
                };
                EnhancedScatterChartWebBehavior.prototype.getMarkerFillOpacity = function (hasSize, shouldEnableFill, hasSelection, isSelected) {
                    if (hasSize || shouldEnableFill) {
                        if (hasSelection && !isSelected) {
                            return this.dimmedBubbleOpacity;
                        }
                        return this.defaultBubbleOpacity;
                    }
                    else {
                        return 0;
                    }
                };
                EnhancedScatterChartWebBehavior.prototype.getMarkerStrokeOpacity = function (hasSize, colorBorder, hasSelection, isSelected) {
                    if (hasSize && colorBorder) {
                        return 1;
                    }
                    else {
                        if (hasSelection && !isSelected) {
                            return this.dimmedBubbleOpacity;
                        }
                        return this.defaultBubbleOpacity;
                    }
                };
                return EnhancedScatterChartWebBehavior;
            }());
            EnhancedScatterChart1443994985041.EnhancedScatterChartWebBehavior = EnhancedScatterChartWebBehavior;
        })(EnhancedScatterChart1443994985041 = visuals.EnhancedScatterChart1443994985041 || (visuals.EnhancedScatterChart1443994985041 = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
var powerbi;
(function (powerbi) {
    var visuals;
    (function (visuals) {
        var plugins;
        (function (plugins) {
            plugins.EnhancedScatterChart1443994985041 = {
                name: 'EnhancedScatterChart1443994985041',
                class: 'EnhancedScatterChart1443994985041',
                capabilities: powerbi.visuals.EnhancedScatterChart1443994985041.EnhancedScatterChart.capabilities,
                custom: true,
                create: function () { return new powerbi.visuals.EnhancedScatterChart1443994985041.EnhancedScatterChart(); }
            };
        })(plugins = visuals.plugins || (visuals.plugins = {}));
    })(visuals = powerbi.visuals || (powerbi.visuals = {}));
})(powerbi || (powerbi = {}));
